from __future__ import annotations
from importlib.metadata import version
import logging
import os
import signal
import subprocess
import threading
import time
from typing import IO, TextIO
from con_duct._models import LogPaths, Outputs, RecordTypes, SessionMode
from con_duct._output import TailPipe, prepare_outputs, remove_files, safe_close_files
from con_duct._signals import SigIntHandler
from con_duct._tracker import Report, monitor_process

__version__ = version("con-duct")

lgr = logging.getLogger("con-duct")

DUCT_OUTPUT_PREFIX = os.getenv("DUCT_OUTPUT_PREFIX", ".duct/logs/{datetime}-{pid}_")
EXECUTION_SUMMARY_FORMAT = (
    "Summary:\n"
    "Exit Code: {exit_code!E}\n"
    "Command: {command}\n"
    "Log files location: {logs_prefix}\n"
    "Wall Clock Time: {wall_clock_time:.3f} sec\n"
    "Memory Peak Usage (RSS): {peak_rss!S}\n"
    "Memory Average Usage (RSS): {average_rss!S}\n"
    "Virtual Memory Peak Usage (VSZ): {peak_vsz!S}\n"
    "Virtual Memory Average Usage (VSZ): {average_vsz!S}\n"
    "Memory Peak Percentage: {peak_pmem:.2f!N}%\n"
    "Memory Average Percentage: {average_pmem:.2f!N}%\n"
    "CPU Peak Usage: {peak_pcpu:.2f!N}%\n"
    "Average CPU Usage: {average_pcpu:.2f!N}%\n"
)


def execute(
    command: str,
    command_args: list[str],
    output_prefix: str,
    sample_interval: float,
    report_interval: float,
    fail_time: float,
    clobber: bool,
    capture_outputs: Outputs,
    outputs: Outputs,
    record_types: RecordTypes,
    summary_format: str,
    colors: bool,
    mode: SessionMode,
    message: str = "",
) -> int:
    """A wrapper to execute a command, monitor and log the process details.

    Returns exit code of the executed process.
    """
    if report_interval < sample_interval:
        raise ValueError(
            "--report-interval must be greater than or equal to --sample-interval."
        )
    if sample_interval < 1.0:
        lgr.warning(
            "--sample-interval=%.3f is below 1.0s and may behave erratically. "
            "See docs/resource-statistics.md for details.",
            sample_interval,
        )

    log_paths = LogPaths.create(output_prefix, pid=os.getpid())
    try:
        log_paths.prepare_paths(clobber, capture_outputs)
    except FileExistsError as e:
        lgr.error("%s", e)
        return 1
    stdout, stderr = prepare_outputs(capture_outputs, outputs, log_paths)
    stdout_file: TextIO | IO[bytes] | int | None
    if isinstance(stdout, TailPipe):
        stdout_file = open(stdout.file_path, "wb")
    else:
        stdout_file = stdout
    stderr_file: TextIO | IO[bytes] | int | None
    if isinstance(stderr, TailPipe):
        stderr_file = open(stderr.file_path, "wb")
    else:
        stderr_file = stderr

    working_directory = os.getcwd()
    full_command = " ".join([str(command)] + command_args)
    files_to_close = [stdout_file, stdout, stderr_file, stderr]

    report = Report(
        command,
        command_args,
        log_paths,
        summary_format,
        working_directory,
        colors,
        clobber,
        message=message,
    )
    files_to_close.append(report.usage_file)

    report.start_time = time.time()
    try:
        report.process = process = subprocess.Popen(
            [str(command)] + command_args,
            stdout=stdout_file,
            stderr=stderr_file,
            start_new_session=(mode == SessionMode.NEW_SESSION),
            cwd=report.working_directory,
        )
    except FileNotFoundError:
        # We failed to execute due to file not found in PATH
        # We should remove log etc files since they are 0-sized
        # degenerates etc
        safe_close_files(files_to_close)
        remove_files(log_paths, assert_empty=True)
        # mimicking behavior of bash and zsh.
        lgr.error("%s: command not found", command)
        return 127  # seems what zsh and bash return then

    signal.signal(signal.SIGINT, SigIntHandler(process.pid))
    lgr.info("duct %s is executing %r...", __version__, full_command)
    lgr.info("Log files will be written to %s", log_paths.prefix)
    try:
        if mode == SessionMode.NEW_SESSION:
            report.session_id = os.getsid(
                process.pid
            )  # Get session ID of the new process
        else:  # CURRENT_SESSION mode
            report.session_id = os.getsid(
                os.getpid()
            )  # Get session ID of duct's own process
    except ProcessLookupError:  # process has already finished
        # TODO: log this at least.
        pass
    stop_event = threading.Event()
    if record_types.has_processes_samples():
        monitoring_args = [
            report,
            process,
            report_interval,
            sample_interval,
            stop_event,
        ]
        monitoring_thread = threading.Thread(
            target=monitor_process, args=monitoring_args
        )
        monitoring_thread.start()
    else:
        monitoring_thread = None

    if record_types.has_system_summary():
        env_thread = threading.Thread(target=report.collect_environment)
        env_thread.start()
        sys_info_thread = threading.Thread(target=report.get_system_info)
        sys_info_thread.start()
    else:
        env_thread, sys_info_thread = None, None

    process.wait()
    report.end_time = time.time()
    lgr.debug("Process ended, setting stop_event to stop monitoring thread")
    stop_event.set()
    if monitoring_thread is not None:
        lgr.debug("Waiting for monitoring thread to finish")
        monitoring_thread.join()
        lgr.debug("Monitoring thread finished")

    # If we have any extra samples that haven't been written yet, do it now
    if report.current_sample is not None:
        report.write_subreport()

    report.process = process
    if env_thread is not None:
        lgr.debug("Waiting for environment collection thread to finish")
        env_thread.join()
        lgr.debug("Environment collection finished")

    if sys_info_thread is not None:
        lgr.debug("Waiting for system information collection thread to finish")
        sys_info_thread.join()
        lgr.debug("System information collection finished")

    if record_types.has_system_summary():
        with open(log_paths.info, "w") as system_logs:
            report.run_time_seconds = f"{report.end_time - report.start_time}"
            system_logs.write(report.dump_json())
    safe_close_files(files_to_close)
    if process.returncode != 0 and (report.elapsed_time < fail_time or fail_time < 0):
        lgr.info(
            "Removing log files since command failed%s.",
            f" in less than {fail_time} seconds" if fail_time > 0 else "",
        )
        remove_files(log_paths)
    else:
        lgr.info(report.execution_summary_formatted)
    return report.process.returncode



from __future__ import annotations
import json
import logging
import multiprocessing
import os
from pathlib import Path
import signal
import subprocess
import sys
from time import sleep, time
import pytest
from utils import assert_files, run_duct_command
from con_duct import _duct_main
from con_duct._constants import SUFFIXES
from con_duct._models import Outputs


def test_sample_less_than_report_interval(temp_output_dir: str) -> None:
    run_duct_command(
        ["echo", "test"],
        sample_interval=0.01,
        report_interval=0.1,
        output_prefix=temp_output_dir,
    )


def test_sample_equal_to_report_interval(temp_output_dir: str) -> None:
    run_duct_command(
        ["echo", "test"],
        sample_interval=0.1,
        report_interval=0.1,
        output_prefix=temp_output_dir,
    )


def test_sample_greater_than_report_interval() -> None:
    with pytest.raises(
        ValueError,
        match="--report-interval must be greater than or equal to --sample-interval",
    ):
        run_duct_command(
            ["echo", "test"],
            sample_interval=1.0,
            report_interval=0.1,
        )


TEST_SCRIPT_DIR = Path(__file__).parent.parent / "data"

expected_files = [
    SUFFIXES["stdout"],
    SUFFIXES["stderr"],
    SUFFIXES["info"],
    SUFFIXES["usage"],
]


def assert_expected_files(temp_output_dir: str, exists: bool = True) -> None:
    assert_files(temp_output_dir, expected_files, exists=exists)


def test_sanity_green(temp_output_dir: str) -> None:
    t0 = time()
    expected_exit_code = 0
    assert (
        run_duct_command(
            ["echo", "hello", "world"],
            sample_interval=4.0,
            report_interval=60.0,
            output_prefix=temp_output_dir,
        )
        == expected_exit_code
    )
    assert time() - t0 < 0.4  # we should not wait for a sample or report interval
    assert_expected_files(temp_output_dir)


@pytest.mark.flaky(reruns=5)
def test_execution_summary(
    request: pytest.FixtureRequest, temp_output_dir: str
) -> None:
    # Scale sleep time on retries to handle slow CI runners (PyPy, Mac)
    attempt = getattr(request.node, "execution_count", 1)
    sleep_time = 0.1 * attempt
    assert (
        run_duct_command(
            ["sleep", str(sleep_time)],
            sample_interval=0.05,  # small enough to ensure we collect at least 1 sample
            report_interval=sleep_time,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    with open(os.path.join(temp_output_dir, SUFFIXES["info"])) as info:
        info_dict = json.loads(info.read())
    execution_summary = info_dict["execution_summary"]
    # Since resources used should be small lets make sure values are roughly sane
    assert execution_summary["average_pmem"] < 10
    assert execution_summary["peak_pmem"] < 10
    assert execution_summary["average_pcpu"] < 10
    assert execution_summary["peak_pcpu"] < 10
    assert execution_summary["exit_code"] == 0
    assert execution_summary["working_directory"] == os.getcwd()


@pytest.mark.parametrize("exit_code", [1, 2, 128])
def test_sanity_red(
    caplog: pytest.LogCaptureFixture, exit_code: int, temp_output_dir: str
) -> None:
    caplog.set_level("INFO")
    assert (
        run_duct_command(
            ["sh", "-c", f"exit {exit_code}"],
            output_prefix=temp_output_dir,
            fail_time=0,  # keep log files regardless of exit code
        )
        == exit_code
    )
    assert f"Exit Code: {exit_code}" in caplog.records[-1].message

    # We still should execute normally
    assert_expected_files(temp_output_dir)


def test_outputs_full(temp_output_dir: str) -> None:
    script_path = str(TEST_SCRIPT_DIR / "test_script.py")
    assert (
        run_duct_command(
            [script_path, "--duration", "1"],
            # It is our default, but let's be explicit
            capture_outputs=Outputs.ALL,
            outputs=Outputs.ALL,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    assert_expected_files(temp_output_dir)


def test_outputs_passthrough(temp_output_dir: str) -> None:
    script_path = str(TEST_SCRIPT_DIR / "test_script.py")
    assert (
        run_duct_command(
            [script_path, "--duration", "1"],
            capture_outputs=Outputs.NONE,
            outputs=Outputs.ALL,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    expected_files = [SUFFIXES["info"], SUFFIXES["usage"]]
    assert_files(temp_output_dir, expected_files, exists=True)
    not_expected_files = [SUFFIXES["stdout"], SUFFIXES["stderr"]]
    assert_files(temp_output_dir, not_expected_files, exists=False)


def test_outputs_capture(temp_output_dir: str) -> None:
    script_path = str(TEST_SCRIPT_DIR / "test_script.py")
    assert (
        run_duct_command(
            [script_path, "--duration", "1"],
            capture_outputs=Outputs.ALL,
            outputs=Outputs.NONE,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    # TODO make this work assert mock.call("this is of test of STDOUT\n") not in mock_stdout.write.mock_calls

    assert_expected_files(temp_output_dir)


def test_outputs_none(temp_output_dir: str) -> None:
    script_path = str(TEST_SCRIPT_DIR / "test_script.py")
    assert (
        run_duct_command(
            [script_path, "--duration", "1"],
            capture_outputs=Outputs.NONE,
            outputs=Outputs.NONE,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    # assert mock.call("this is of test of STDOUT\n") not in mock_stdout.write.mock_calls

    expected_files = [SUFFIXES["info"], SUFFIXES["usage"]]
    assert_files(temp_output_dir, expected_files, exists=True)

    not_expected_files = [SUFFIXES["stdout"], SUFFIXES["stderr"]]
    assert_files(temp_output_dir, not_expected_files, exists=False)


def test_exit_before_first_sample(temp_output_dir: str) -> None:
    assert (
        run_duct_command(
            ["ls"],
            sample_interval=0.1,
            report_interval=0.1,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    assert_expected_files(temp_output_dir)
    # TODO check usagefile


def test_run_less_than_report_interval(temp_output_dir: str) -> None:
    assert (
        run_duct_command(
            ["sleep", "0.01"],
            sample_interval=0.001,
            report_interval=0.1,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    # Specifically we need to assert that usage file gets written anyway.
    assert_expected_files(temp_output_dir)


def test_execute_unknown_command(
    temp_output_dir: str, caplog: pytest.LogCaptureFixture
) -> None:
    cmd = "this_command_does_not_exist_123abrakadabra"
    with caplog.at_level(logging.ERROR):
        assert run_duct_command([cmd]) == 127
    assert f"{cmd}: command not found" in caplog.text
    assert_expected_files(temp_output_dir, exists=False)


def _runner_for_signal_int(temp_output_dir: str, fail_time: float | None) -> int:
    kws = {}
    if fail_time is not None:
        kws["fail_time"] = fail_time
    return run_duct_command(
        ["sleep", "60.74016230000801"], output_prefix=temp_output_dir, **kws
    )


@pytest.mark.flaky(reruns=5)
@pytest.mark.parametrize("fail_time", [None, 0, 10, -1, -3.14])
def test_signal_int(
    request: pytest.FixtureRequest, temp_output_dir: str, fail_time: float | None
) -> None:
    # Scale wait time on retries to handle slow CI runners (PyPy, Mac)
    attempt = getattr(request.node, "execution_count", 1)
    wait_time = 0.2 * attempt
    proc = multiprocessing.Process(
        target=_runner_for_signal_int, args=(temp_output_dir, fail_time)
    )
    proc.start()
    sleep(wait_time)
    assert proc.pid is not None, "Process PID should not be None"  # for mypy
    os.kill(proc.pid, signal.SIGINT)
    proc.join()

    # Once the command has been killed, duct should exit gracefully with exit code 0
    assert proc.exitcode == 0

    if fail_time is None or fail_time != 0:
        assert_expected_files(temp_output_dir, exists=False)
    else:
        # proc exit code should Cannot retrieve the exit code from the thread, it is written to the file
        with open(os.path.join(temp_output_dir, SUFFIXES["info"])) as info:
            info_data = json.loads(info.read())

        command_exit_code = info_data["execution_summary"]["exit_code"]
        # SIGINT
        assert command_exit_code == 128 + 2


def _runner_for_signal_kill(temp_output_dir: str, fail_time: float | None) -> int:
    script_path = str(TEST_SCRIPT_DIR / "signal_ignorer.py")
    kws = {}
    if fail_time is not None:
        kws["fail_time"] = fail_time
    return run_duct_command([script_path], output_prefix=temp_output_dir, **kws)


@pytest.mark.flaky(reruns=5)
@pytest.mark.parametrize("fail_time", [None, 0, 10, -1, -3.14])
def test_signal_kill(
    request: pytest.FixtureRequest, temp_output_dir: str, fail_time: float | None
) -> None:
    # Scale wait time on retries to handle slow CI runners (PyPy, Mac)
    attempt = getattr(request.node, "execution_count", 1)
    wait_time = 0.2 * attempt
    proc = multiprocessing.Process(
        target=_runner_for_signal_kill, args=(temp_output_dir, fail_time)
    )
    proc.start()
    sleep(wait_time)
    assert proc.pid is not None, "Process PID should not be None"  # for mypy
    os.kill(proc.pid, signal.SIGINT)
    sleep(wait_time)
    os.kill(proc.pid, signal.SIGINT)
    sleep(wait_time)
    os.kill(proc.pid, signal.SIGINT)
    proc.join()

    # Once the command has been killed, duct should exit gracefully with exit code 0
    assert proc.exitcode == 0

    if fail_time is None or fail_time != 0:
        assert_expected_files(temp_output_dir, exists=False)
    else:
        # Cannot retrieve the command exit code from the thread, get from duct log
        with open(os.path.join(temp_output_dir, SUFFIXES["info"])) as info:
            info_data = json.loads(info.read())

        command_exit_code = info_data["execution_summary"]["exit_code"]
        # SIGKILL
        assert command_exit_code == 128 + 9


def test_duct_as_executable(temp_output_dir: str) -> None:
    ps_command = [
        sys.executable,
        _duct_main.__file__,
        "-p",
        temp_output_dir,
        "-q",
        "sleep",
        "0.01",
    ]
    # Assert does not raise
    subprocess.check_output(ps_command, shell=False).decode()


def test_message_in_json_output(temp_output_dir: str) -> None:
    """Test that message appears in JSON output when provided."""
    test_message = "Electrolytes, its what plants crave"
    assert (
        run_duct_command(
            ["echo", "hello"],
            output_prefix=temp_output_dir,
            message=test_message,
        )
        == 0
    )

    # Check that message appears in info.json
    with open(os.path.join(temp_output_dir, SUFFIXES["info"])) as info:
        info_dict = json.loads(info.read())

    assert "message" in info_dict
    assert info_dict["message"] == test_message


def test_no_message_in_json_output(temp_output_dir: str) -> None:
    """Test that message field is empty string when not provided."""
    assert (
        run_duct_command(
            ["echo", "hello"],
            output_prefix=temp_output_dir,
        )
        == 0
    )

    # Check that message field is empty string in info.json
    with open(os.path.join(temp_output_dir, SUFFIXES["info"])) as info:
        info_dict = json.loads(info.read())

    assert "message" in info_dict
    assert info_dict["message"] == ""


def test_execute_existing_files_no_clobber(
    temp_output_dir: str, caplog: pytest.LogCaptureFixture
) -> None:
    Path(temp_output_dir, SUFFIXES["info"]).write_text("pre-existing")
    with caplog.at_level(logging.ERROR):
        assert run_duct_command(["echo", "hi"], output_prefix=temp_output_dir) == 1
    assert "Conflicting files:" in caplog.text
    assert_files(
        temp_output_dir,
        [SUFFIXES["stdout"], SUFFIXES["stderr"], SUFFIXES["usage"]],
        exists=False,
    )



import argparse
import contextlib
from io import StringIO
import json
import logging
import os
import tempfile
from typing import Any, Dict, Optional
import unittest
from unittest.mock import mock_open, patch
import pytest
from con_duct._constants import __schema_version__
from con_duct._formatter import SummaryFormatter
from con_duct.ls import (
    MINIMUM_SCHEMA_VERSION,
    _flatten_dict,
    _restrict_row,
    ensure_compliant_schema,
    load_duct_runs,
    ls,
    process_run_data,
)


def test_load_duct_runs_sanity() -> None:
    mock_json = json.dumps(
        {"schema_version": "0.2.1", "prefix": "/test/path_", "command": "echo hello"}
    )
    with patch("builtins.open", mock_open(read_data=mock_json)):
        result = load_duct_runs(["/test/path_info.json"])
    assert len(result) == 1
    assert result[0]["prefix"] == "/test/path_"


def test_load_duct_runs_skips_unsupported_schema() -> None:
    mock_json = json.dumps(
        {"schema_version": "0.1.1", "prefix": "/test/path_", "command": "echo hello"}
    )
    with patch("builtins.open", mock_open(read_data=mock_json)):
        result = load_duct_runs(["/test/path_info.json"])
    assert len(result) == 0


def test_load_duct_runs_uses_filenames_not_stored_prefix() -> None:
    mock_json = json.dumps(
        {
            "schema_version": "0.2.1",
            "prefix": "/test/not_anymore_",
            "command": "echo hello",
        }
    )
    with patch("builtins.open", mock_open(read_data=mock_json)):
        result = load_duct_runs(["/actual_filepath_info.json"])
    assert len(result) == 1
    assert result[0]["prefix"] == "/actual_filepath_"


def test_flatten_dict() -> None:
    nested = {"a": {"b": 1, "c": 2}, "d": 3}
    result = _flatten_dict(nested)
    assert result == {"b": 1, "c": 2, "d": 3}


def test_restrict_row() -> None:
    row = {"prefix": "/test/path", "exit_code": 0, "extra": "ignore"}
    fields = ["exit_code"]
    result = _restrict_row(fields, row)
    assert "prefix" in result
    assert "exit_code" in result
    assert "extra" not in result


def test_process_run_data() -> None:
    run_data = [
        {
            "prefix": "/test/path",
            "exit_code": 0,
            "wall_clock_time": 0.12345678,
        }
    ]
    formatter = SummaryFormatter(enable_colors=False)
    result = process_run_data(run_data, ["wall_clock_time"], formatter)
    assert isinstance(result, list)
    assert result[0]["prefix"] == "/test/path"
    assert "exit_code" not in result[0]
    assert result[0]["wall_clock_time"] == "0.123 sec"


def test_ensure_compliant_schema_noop_for_current_version() -> None:
    info: Dict[str, Any] = {
        "schema_version": __schema_version__,
        "execution_summary": {},
    }
    ensure_compliant_schema(info)
    assert "working_directory" not in info["execution_summary"]


def test_ensure_compliant_schema_adds_field_for_old_version() -> None:
    info: Dict[str, Any] = {"schema_version": "0.2.0", "execution_summary": {}}
    ensure_compliant_schema(info)
    assert info["execution_summary"]["working_directory"] == ""
    assert info["message"] == ""


def test_ensure_compliant_schema_ignores_unexpected_future_version() -> None:
    info: Dict[str, Any] = {"schema_version": "99.0.0", "execution_summary": {}}
    ensure_compliant_schema(info)
    assert "working_directory" not in info["execution_summary"]


def test_load_duct_runs_handles_empty_json_files(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test desired behavior: empty JSON files produce debug logs and are skipped."""
    with patch("builtins.open", mock_open(read_data="")):
        with caplog.at_level(logging.DEBUG):
            result = load_duct_runs(["/test/empty_info.json"])

    # empty files result in empty list
    assert len(result) == 0
    # empty files result in debug level log (not warning)
    assert len([r for r in caplog.records if r.levelname == "DEBUG"]) == 1
    assert not any(r for r in caplog.records if r.levelname == "WARNING")
    assert "Skipping empty file" in caplog.text


def test_load_duct_runs_handles_invalid_json_files(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test current behavior: invalid JSON files produce warnings and are skipped."""
    with patch("builtins.open", mock_open(read_data="not json at all")):
        with caplog.at_level(logging.WARNING):
            result = load_duct_runs(["/test/invalid_info.json"])

    assert len(result) == 0
    assert len(caplog.records) == 1
    assert "Failed to load file" in caplog.text


def test_load_duct_runs_mixed_empty_and_valid_files(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test behavior with mix of empty and valid JSON files."""
    valid_json = json.dumps(
        {"schema_version": "0.2.1", "prefix": "/test/path_", "command": "echo hello"}
    )

    def side_effect(filename: str) -> Any:
        if "empty" in filename:
            return mock_open(read_data="")()
        else:
            return mock_open(read_data=valid_json)()

    with patch("builtins.open", side_effect=side_effect):
        with caplog.at_level(logging.DEBUG):
            result = load_duct_runs(["/test/empty_info.json", "/test/valid_info.json"])

    # only valid file is loaded
    assert len(result) == 1
    assert result[0]["prefix"] == "/test/valid_"
    # debug log for empty file, no warning
    assert len([r for r in caplog.records if r.levelname == "DEBUG"]) == 1
    assert not any(r for r in caplog.records if r.levelname == "WARNING")
    assert "Skipping empty file" in caplog.text


class TestLS(unittest.TestCase):
    def setUp(self) -> None:
        """Create a temporary directory and test files."""
        self.temp_dir = tempfile.TemporaryDirectory()
        self.old_cwd = os.getcwd()
        os.chdir(self.temp_dir.name)
        self.files = {
            "file1_info.json": {
                "schema_version": MINIMUM_SCHEMA_VERSION,
                "execution_summary": {},
                "prefix": "test1",
                "filter_this": "yes",
            },
            "file2_info.json": {
                "schema_version": MINIMUM_SCHEMA_VERSION,
                "execution_summary": {},
                "prefix": "test2",
                "filter_this": "no",
            },
            "file3_info.json": {
                "schema_version": "0.1.0",
                "execution_summary": {},
                "prefix": "old_version",
            },
            "not_matching.json": {
                "schema_version": MINIMUM_SCHEMA_VERSION,
                "execution_summary": {},
                "prefix": "no_match",
            },
            ".duct/logs/default_logpath_info.json": {
                "schema_version": MINIMUM_SCHEMA_VERSION,
                "execution_summary": {},
                "prefix": "default_file1",
            },
        }
        for filename, content in self.files.items():
            full_path = os.path.join(self.temp_dir.name, filename)
            os.makedirs(os.path.dirname(full_path), exist_ok=True)
            with open(full_path, "w") as f:
                json.dump(content, f)

    def tearDown(self) -> None:
        """Clean up the temporary directory."""
        os.chdir(self.old_cwd)
        self.temp_dir.cleanup()

    def _run_ls(
        self, paths: list[str], fmt: str, args: Optional[argparse.Namespace] = None
    ) -> str:
        """Helper function to run ls() and capture stdout."""
        if args is None:
            args = argparse.Namespace(
                paths=[os.path.join(self.temp_dir.name, path) for path in paths],
                colors=False,
                fields=["prefix", "schema_version"],
                eval_filter=None,
                format=fmt,
                func=ls,
                reverse=False,
            )
        buf = StringIO()
        with contextlib.redirect_stdout(buf):
            exit_code = ls(args)
            assert exit_code == 0
        return buf.getvalue().strip()

    def test_ls_sanity(self) -> None:
        """Basic sanity test to ensure ls() runs without crashing."""
        just_file1 = ["file1_info.json"]
        result = self._run_ls(just_file1, "summaries")

        assert "Prefix:" in result
        prefixes = [
            line.split(":", 1)[1].strip()
            for line in result.splitlines()
            if line.startswith("Prefix:")
        ]
        assert len(prefixes) == 1
        assert any("file1" in p for p in prefixes)

    def test_ls_with_filter(self) -> None:
        """Basic sanity test to ensure ls() runs without crashing."""
        paths = ["file1_info.json", "file2_info.json"]
        args = argparse.Namespace(
            paths=[os.path.join(self.temp_dir.name, path) for path in paths],
            colors=False,
            fields=["prefix", "schema_version"],
            eval_filter="filter_this=='yes'",
            format="summaries",
            func=ls,
            reverse=False,
        )
        result = self._run_ls(paths, "summaries", args)

        assert "Prefix:" in result
        prefixes = [
            line.split(":", 1)[1].strip()
            for line in result.splitlines()
            if line.startswith("Prefix:")
        ]
        assert len(prefixes) == 1
        assert any("file1" in p for p in prefixes)
        # filter_this == 'no'
        assert "file2" not in result

    def test_ls_no_pos_args(self) -> None:
        result = self._run_ls([], "summaries")

        assert "Prefix:" in result
        prefixes = [
            line.split(":", 1)[1].strip()
            for line in result.splitlines()
            if line.startswith("Prefix:")
        ]
        assert len(prefixes) == 1
        assert any("default_logpath" in p for p in prefixes)

        assert "file1" not in result
        assert "file2" not in result
        assert "file3" not in result
        assert "not_matching.json" not in result

    def test_ls_multiple_paths(self) -> None:
        """Basic sanity test to ensure ls() runs without crashing."""
        files_1_and_2 = ["file1_info.json", "file2_info.json"]
        result = self._run_ls(files_1_and_2, "summaries")

        assert "Prefix:" in result
        prefixes = [
            line.split(":", 1)[1].strip()
            for line in result.splitlines()
            if line.startswith("Prefix:")
        ]
        assert len(prefixes) == 2
        assert any("file1" in p for p in prefixes)
        assert any("file2" in p for p in prefixes)

    def test_ls_ignore_old_schema(self) -> None:
        """Basic sanity test to ensure ls() runs without crashing."""
        files_1_2_3 = ["file1_info.json", "file2_info.json", "file3_info.json"]
        result = self._run_ls(files_1_2_3, "summaries")

        assert "Prefix:" in result
        prefixes = [
            line.split(":", 1)[1].strip()
            for line in result.splitlines()
            if line.startswith("Prefix:")
        ]
        assert len(prefixes) == 2
        assert any("file1" in p for p in prefixes)
        assert any("file2" in p for p in prefixes)
        # file3 does not meet minimum schema version
        assert "file3" not in result

    def test_ls_ignore_non_infojson(self) -> None:
        """Basic sanity test to ensure ls() runs without crashing."""
        all_files = ["file1_info.json", "file2_info.json", "not_matching.json"]
        result = self._run_ls(all_files, "summaries")

        assert "Prefix:" in result
        prefixes = [
            line.split(":", 1)[1].strip()
            for line in result.splitlines()
            if line.startswith("Prefix:")
        ]
        assert len(prefixes) == 2
        assert any("file1" in p for p in prefixes)
        assert any("file2" in p for p in prefixes)
        # does not end in info.json
        assert "not_matching.json" not in result

    def test_ls_json_output(self) -> None:
        """Test JSON output format."""
        result = self._run_ls(["file1_info.json"], "json")
        parsed = json.loads(result)
        assert len(parsed) == 1
        assert "prefix" in parsed[0]

    def test_ls_json_pp_output(self) -> None:
        """Test pretty-printed JSON output format."""
        result = self._run_ls(["file1_info.json"], "json_pp")
        parsed = json.loads(result)
        assert len(parsed) == 1
        assert "prefix" in parsed[0]

    def test_ls_yaml_output(self) -> None:
        """Test YAML output format."""
        yaml = pytest.importorskip("yaml")
        result = self._run_ls(["file1_info.json"], "yaml")
        parsed = yaml.safe_load(result)
        assert len(parsed) == 1
        assert "prefix" in parsed[0]

    def test_ls_pyout_output(self) -> None:
        """Test pyout output format."""
        pytest.importorskip("pyout")
        result = self._run_ls(["file1_info.json"], "pyout")
        # pyout header
        assert "PREFIX" in result
        assert os.path.join(self.temp_dir.name, "file1_") in result

    def test_ls_reverse(self) -> None:
        """Test --reverse flag lists entries in reverse order."""
        paths = ["file1_info.json", "file2_info.json"]

        # Get normal order
        result_normal = self._run_ls(paths, "json")
        parsed_normal = json.loads(result_normal)
        prefixes_normal = [row["prefix"] for row in parsed_normal]

        # Get reversed order
        args = argparse.Namespace(
            paths=[os.path.join(self.temp_dir.name, path) for path in paths],
            colors=False,
            fields=["prefix", "schema_version"],
            eval_filter=None,
            format="json",
            func=ls,
            reverse=True,
        )
        result_reversed = self._run_ls(paths, "json", args)
        parsed_reversed = json.loads(result_reversed)
        prefixes_reversed = [row["prefix"] for row in parsed_reversed]

        assert prefixes_reversed == list(reversed(prefixes_normal))



from __future__ import annotations
import logging
import os
from pathlib import Path
import sys
from unittest.mock import patch
import pytest
from con_duct import cli


@pytest.fixture
def temp_env_files(tmp_path: Path) -> dict[str, Path]:
    """Create temporary .env files for testing."""
    # System-level config
    system_env = tmp_path / "system.env"
    system_env.write_text("DUCT_LOG_LEVEL=WARNING\nDUCT_SAMPLE_INTERVAL=2.0\n")

    # User-level config
    user_env = tmp_path / "user.env"
    user_env.write_text("DUCT_LOG_LEVEL=DEBUG\nDUCT_REPORT_INTERVAL=120.0\n")

    # Project-level config
    project_env = tmp_path / "project.env"
    project_env.write_text("DUCT_LOG_LEVEL=INFO\nDUCT_MESSAGE=test message\n")

    return {
        "system": system_env,
        "user": user_env,
        "project": project_env,
    }


def test_load_env_files_basic(
    temp_env_files: dict[str, Path], clean_env: pytest.MonkeyPatch
) -> None:
    """Test basic .env file loading."""
    pytest.importorskip("dotenv")

    config_paths = f"{temp_env_files['system']}{os.pathsep}{temp_env_files['user']}"
    clean_env.setenv("DUCT_CONFIG_PATHS", config_paths)

    cli.load_duct_env_files()
    # User config should override system config
    assert os.environ.get("DUCT_LOG_LEVEL") == "DEBUG"
    assert os.environ.get("DUCT_SAMPLE_INTERVAL") == "2.0"
    assert os.environ.get("DUCT_REPORT_INTERVAL") == "120.0"


def test_load_env_files_precedence(
    temp_env_files: dict[str, Path], clean_env: pytest.MonkeyPatch
) -> None:
    """Test that later files override earlier files."""
    pytest.importorskip("dotenv")

    config_paths = os.pathsep.join(
        [
            str(temp_env_files["system"]),
            str(temp_env_files["user"]),
            str(temp_env_files["project"]),
        ]
    )
    clean_env.setenv("DUCT_CONFIG_PATHS", config_paths)

    cli.load_duct_env_files()
    # Project should override user which overrides system
    assert os.environ.get("DUCT_LOG_LEVEL") == "INFO"
    assert os.environ.get("DUCT_MESSAGE") == "test message"
    # Values from earlier files that weren't overridden
    assert os.environ.get("DUCT_SAMPLE_INTERVAL") == "2.0"
    assert os.environ.get("DUCT_REPORT_INTERVAL") == "120.0"


def test_explicit_env_vars_win(
    temp_env_files: dict[str, Path], clean_env: pytest.MonkeyPatch
) -> None:
    """Test that explicit environment variables are not overridden by .env files."""
    pytest.importorskip("dotenv")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(temp_env_files["project"]))
    clean_env.setenv("DUCT_LOG_LEVEL", "CRITICAL")

    cli.load_duct_env_files()
    # Explicit env var should NOT be overridden
    assert os.environ.get("DUCT_LOG_LEVEL") == "CRITICAL"
    # But .env file should still set other vars
    assert os.environ.get("DUCT_MESSAGE") == "test message"


def test_missing_env_file_ignored(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that missing .env files are silently ignored."""
    pytest.importorskip("dotenv")

    nonexistent = tmp_path / "nonexistent.env"
    clean_env.setenv("DUCT_CONFIG_PATHS", str(nonexistent))

    # Should not raise an exception
    cli.load_duct_env_files()


def test_xdg_config_home_expansion(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that ${XDG_CONFIG_HOME:-~/.config} syntax is expanded correctly."""
    pytest.importorskip("dotenv")

    # Create a .env file in a custom XDG location
    xdg_dir = tmp_path / "xdg_config"
    xdg_dir.mkdir()
    duct_dir = xdg_dir / "duct"
    duct_dir.mkdir()
    env_file = duct_dir / ".env"
    env_file.write_text("DUCT_LOG_LEVEL=ERROR\n")

    clean_env.setenv("XDG_CONFIG_HOME", str(xdg_dir))
    clean_env.setenv("DUCT_CONFIG_PATHS", "${XDG_CONFIG_HOME:-~/.config}/duct/.env")

    cli.load_duct_env_files()
    assert os.environ.get("DUCT_LOG_LEVEL") == "ERROR"


def test_multiline_values(tmp_path: Path, clean_env: pytest.MonkeyPatch) -> None:
    """Test that multiline values in .env files are handled correctly."""
    pytest.importorskip("dotenv")

    # Create .env file with multiline value
    env_file = tmp_path / "multiline.env"
    env_file.write_text('DUCT_MESSAGE="Line 1\nLine 2\nLine 3"\n')

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    cli.load_duct_env_files()
    message = os.environ.get("DUCT_MESSAGE")
    assert message is not None
    assert "Line 1" in message
    assert "Line 2" in message
    assert "Line 3" in message


def test_without_python_dotenv(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test graceful degradation when python-dotenv is not installed."""
    # Setting module to None in sys.modules causes ImportError on import
    monkeypatch.setitem(sys.modules, "dotenv", None)

    log_buffer = cli.load_duct_env_files()

    # Should have a message about dotenv not being installed
    messages = [msg for level, msg in log_buffer]
    assert any("python-dotenv not installed" in msg for msg in messages)


def test_early_logging_buffer(tmp_path: Path, clean_env: pytest.MonkeyPatch) -> None:
    """Test that log messages are buffered during .env loading."""
    pytest.importorskip("dotenv")

    # Create a test .env file
    env_file = tmp_path / "test.env"
    env_file.write_text("TEST_VAR=test_value\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    log_buffer = cli.load_duct_env_files()

    # Check that messages were buffered
    assert len(log_buffer) > 0
    # Should have messages about searching and loading
    messages = [msg for level, msg in log_buffer]
    assert any("Searching for .env files" in msg for msg in messages)
    assert any("Loading .env file" in msg for msg in messages)


def test_early_logging_replay(
    tmp_path: Path, clean_env: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    """Test that buffered messages are replayed through the logger."""
    pytest.importorskip("dotenv")

    # Create a test .env file
    env_file = tmp_path / "test.env"
    env_file.write_text("TEST_VAR=test_value\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    log_buffer = cli.load_duct_env_files()

    # Buffer should have messages
    assert len(log_buffer) > 0

    # Configure logging and replay
    with caplog.at_level(logging.DEBUG, logger="con-duct"):
        cli._replay_early_logs(log_buffer)

    # Check that messages were logged
    assert len(caplog.records) > 0
    logged_messages = [record.message for record in caplog.records]
    assert any("Searching for .env files" in msg for msg in logged_messages)


def test_early_logging_respects_level(
    tmp_path: Path, clean_env: pytest.MonkeyPatch, caplog: pytest.LogCaptureFixture
) -> None:
    """Test that replayed messages respect the configured log level."""
    pytest.importorskip("dotenv")

    # Create a test .env file
    env_file = tmp_path / "test.env"
    env_file.write_text("TEST_VAR=test_value\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    log_buffer = cli.load_duct_env_files()

    # Buffer should have both DEBUG and INFO messages
    levels = {level for level, _ in log_buffer}
    assert "DEBUG" in levels  # "Searching for .env files"
    assert "INFO" in levels  # "Loading .env file"

    # Replay at WARNING level - neither DEBUG nor INFO should appear
    with caplog.at_level(logging.WARNING, logger="con-duct"):
        cli._replay_early_logs(log_buffer)

    assert len(caplog.records) == 0

    # Replay at INFO level - only INFO messages should appear
    caplog.clear()
    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    log_buffer2 = cli.load_duct_env_files()

    with caplog.at_level(logging.INFO, logger="con-duct"):
        cli._replay_early_logs(log_buffer2)

    assert len(caplog.records) > 0
    assert all(r.levelno >= logging.INFO for r in caplog.records)


def test_permission_denied_handling(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that permission denied errors are handled gracefully."""
    dotenv = pytest.importorskip("dotenv")

    # Create a .env file that exists (so it passes the exists() check)
    env_file = tmp_path / "unreadable.env"
    env_file.write_text("DUCT_LOG_LEVEL=DEBUG\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    # Mock load_dotenv to raise PermissionError (works even as root)
    with patch.object(
        dotenv, "load_dotenv", side_effect=PermissionError("Permission denied")
    ):
        # Should not raise an exception
        log_buffer = cli.load_duct_env_files()

    # Should have a WARNING message about permission denied
    messages = [msg for level, msg in log_buffer]
    assert any("Cannot read .env file" in msg for msg in messages)
    assert any("Permission denied" in msg for msg in messages)


def test_malformed_env_file_handling(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that malformed .env files (null bytes) are handled gracefully."""
    pytest.importorskip("dotenv")

    # Create a .env file with null bytes
    env_file = tmp_path / "malformed.env"
    env_file.write_bytes(b"DUCT_MESSAGE=before\x00after\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    # Should not raise an exception
    log_buffer = cli.load_duct_env_files()

    # Should have a WARNING message about malformed file
    messages = [msg for level, msg in log_buffer]
    assert any("Skipping malformed .env file" in msg for msg in messages)
    # The error message varies by python-dotenv version, so just check it's there
    assert any(level == "WARNING" for level, _ in log_buffer)


def test_invalid_utf8_env_file_handling(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that .env files with invalid UTF-8 are handled gracefully."""
    pytest.importorskip("dotenv")

    # Create a .env file with invalid UTF-8 bytes
    env_file = tmp_path / "bad_encoding.env"
    env_file.write_bytes(b"DUCT_MESSAGE=hello\xff\xfeworld\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    # Should not raise an exception
    log_buffer = cli.load_duct_env_files()

    # Should have a WARNING message about malformed file
    messages = [msg for level, msg in log_buffer]
    assert any("Skipping malformed .env file" in msg for msg in messages)
    assert any("utf-8" in msg.lower() or "unicode" in msg.lower() for msg in messages)


def test_default_config_paths_used(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that DEFAULT_CONFIG_PATHS is used when DUCT_CONFIG_PATHS is not set."""
    pytest.importorskip("dotenv")

    # Create a .env file at the project-level default path (.duct/.env)
    duct_dir = tmp_path / ".duct"
    duct_dir.mkdir()
    env_file = duct_dir / ".env"
    env_file.write_text("DUCT_TEST_DEFAULT_PATH=found\n")

    # Patch DEFAULT_CONFIG_PATHS to use our temp path
    clean_env.setattr(cli, "DEFAULT_CONFIG_PATHS", str(env_file))
    # Ensure DUCT_CONFIG_PATHS is not set (clean_env already handles this)

    cli.load_duct_env_files()
    assert os.environ.get("DUCT_TEST_DEFAULT_PATH") == "found"


def test_bare_variable_expansion(tmp_path: Path, clean_env: pytest.MonkeyPatch) -> None:
    """Test that ${VAR} without default is expanded correctly."""
    pytest.importorskip("dotenv")

    # Create a .env file
    env_file = tmp_path / "test.env"
    env_file.write_text("DUCT_TEST_BARE_VAR=found\n")

    # Set up path using bare ${VAR} syntax
    clean_env.setenv("MY_CONFIG_DIR", str(tmp_path))
    clean_env.setenv("DUCT_CONFIG_PATHS", "${MY_CONFIG_DIR}/test.env")

    cli.load_duct_env_files()
    assert os.environ.get("DUCT_TEST_BARE_VAR") == "found"


def test_empty_config_paths(clean_env: pytest.MonkeyPatch) -> None:
    """Test that empty DUCT_CONFIG_PATHS is handled gracefully."""
    pytest.importorskip("dotenv")

    clean_env.setenv("DUCT_CONFIG_PATHS", "")

    # Should not raise an exception
    log_buffer = cli.load_duct_env_files()

    # Should have logged that no files were found
    messages = [msg for level, msg in log_buffer]
    assert any("No .env files found" in msg for msg in messages)


def test_tilde_expansion(tmp_path: Path, clean_env: pytest.MonkeyPatch) -> None:
    """Test that ~ is expanded to home directory in paths."""
    pytest.importorskip("dotenv")

    # Create a .env file
    env_file = tmp_path / "test.env"
    env_file.write_text("DUCT_TEST_TILDE=found\n")

    # Temporarily change HOME to our temp directory
    clean_env.setenv("HOME", str(tmp_path))
    clean_env.setenv("DUCT_CONFIG_PATHS", "~/test.env")

    cli.load_duct_env_files()
    assert os.environ.get("DUCT_TEST_TILDE") == "found"


def test_env_values_flow_to_argparse(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that .env values are picked up by argparse defaults."""
    pytest.importorskip("dotenv")

    # Create a .env file with duct settings
    env_file = tmp_path / "test.env"
    env_file.write_text("DUCT_SAMPLE_INTERVAL=99.5\nDUCT_REPORT_INTERVAL=300.0\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    # Load .env files (normally called at start of main())
    cli.load_duct_env_files()

    # Now parse args - the defaults should come from env vars set by .env
    args = cli._create_run_parser().parse_args(["echo", "test"])

    assert args.sample_interval == 99.5
    assert args.report_interval == 300.0


def test_env_log_level_flows_to_argparse(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that DUCT_LOG_LEVEL from .env is picked up by the common parser."""
    pytest.importorskip("dotenv")

    # Create a .env file with log level
    env_file = tmp_path / "test.env"
    env_file.write_text("DUCT_LOG_LEVEL=DEBUG\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    # Load .env files
    cli.load_duct_env_files()

    # Parse using main() flow - log_level is in the common parser
    args = cli._create_common_parser().parse_args([])

    assert args.log_level == "DEBUG"


def test_whitespace_in_config_paths(
    tmp_path: Path, clean_env: pytest.MonkeyPatch
) -> None:
    """Test that paths with spaces are handled correctly."""
    pytest.importorskip("dotenv")

    # Create a directory with spaces in the name
    spaced_dir = tmp_path / "path with spaces"
    spaced_dir.mkdir()
    env_file = spaced_dir / "test.env"
    env_file.write_text("DUCT_TEST_SPACES=found\n")

    clean_env.setenv("DUCT_CONFIG_PATHS", str(env_file))

    cli.load_duct_env_files()
    assert os.environ.get("DUCT_TEST_SPACES") == "found"



"""Tests for utility functions in _duct_main.py"""

import pytest
from con_duct._utils import (
    assert_num,
    etime_to_etimes,
    is_same_pid,
    pdcpu_from_pcpu,
)


@pytest.mark.parametrize("input_value", [0, 1, 2, -1, 100, 0.001, -1.68])
def test_assert_num_green(input_value: int) -> None:
    assert_num(input_value)


@pytest.mark.parametrize("input_value", ["hi", "0", "one"])
def test_assert_num_red(input_value: int) -> None:
    with pytest.raises(AssertionError):
        assert_num(input_value)


@pytest.mark.parametrize(
    "etime,expected",
    [
        ("00:42", 42.0),
        ("01:30", 90.0),
        ("01:00:00", 3600.0),
        ("12:34:56", 12 * 3600 + 34 * 60 + 56),
        ("02-03:04:05", 2 * 86400 + 3 * 3600 + 4 * 60 + 5),
        ("100-00:00:00", 100 * 86400.0),
    ],
)
def test_etime_to_etimes_green(etime: str, expected: float) -> None:
    assert etime_to_etimes(etime) == expected


@pytest.mark.parametrize("etime", ["", "garbage", "12", "1:2:3:4", "ab:cd"])
def test_etime_to_etimes_red(etime: str) -> None:
    with pytest.raises(ValueError):
        etime_to_etimes(etime)


@pytest.mark.parametrize(
    "prev_pcpu,prev_etimes,curr_pcpu,curr_etimes,expected",
    [
        # Motivating con/duct#399 case: 100% for 60s then idle for 60s.
        # Lifetime pcpu still reads 50% at t=120 (60 cputime / 120
        # etime), but the corrected reading is 0%.
        (100.0, 60.0, 50.0, 120.0, 0.0),
        # Constant 84% load across a 10s interval.
        (80.0, 10.0, 82.0, 20.0, 84.0),
        # Pid that ramps from 50% lifetime to 75% lifetime over 100s
        # of new wall time -> 100% during the new interval.
        (50.0, 100.0, 75.0, 200.0, 100.0),
    ],
)
def test_pdcpu_from_pcpu_green(
    prev_pcpu: float,
    prev_etimes: float,
    curr_pcpu: float,
    curr_etimes: float,
    expected: float,
) -> None:
    assert pdcpu_from_pcpu(prev_pcpu, prev_etimes, curr_pcpu, curr_etimes) == expected


@pytest.mark.parametrize(
    "prev_pcpu,prev_etimes,curr_pcpu,curr_etimes",
    [
        # etimes regressed -> suspected pid reuse; defensive guard.
        (80.0, 100.0, 10.0, 2.0),
        # Same instant -> interval is zero, no rate definable.
        (50.0, 100.0, 50.0, 100.0),
    ],
)
def test_pdcpu_from_pcpu_returns_none_when_no_interval(
    prev_pcpu: float,
    prev_etimes: float,
    curr_pcpu: float,
    curr_etimes: float,
) -> None:
    assert pdcpu_from_pcpu(prev_pcpu, prev_etimes, curr_pcpu, curr_etimes) is None


@pytest.mark.parametrize(
    "prev_pcpu,prev_etimes,curr_pcpu,curr_etimes",
    [
        # Aggregation-timing skew: max pcpu in prev was captured early in
        # the prev interval (pcpu_max=400 on a 4-core box, then idle).
        # Curr's max happens late, so curr's pcpu*etime estimate of cputime
        # is much smaller than prev's. Δcputime goes negative even though
        # the pid is continuous (Δetime ≈ Δwall).
        (400.0, 59.0, 100.0, 119.0),
        # Smaller version of the same effect.
        (50.0, 60.0, 5.0, 120.0),
    ],
)
def test_pdcpu_from_pcpu_returns_none_on_negative_result(
    prev_pcpu: float,
    prev_etimes: float,
    curr_pcpu: float,
    curr_etimes: float,
) -> None:
    """Negative pdcpu == aggregation-timing artifact (not pid reuse).

    See is_same_pid for identity; this clamp catches the residual case
    where identity holds but the max-vs-end-etime mismatch in duct's
    aggregated records produces a spurious negative delta.
    """
    assert pdcpu_from_pcpu(prev_pcpu, prev_etimes, curr_pcpu, curr_etimes) is None


@pytest.mark.parametrize(
    "prev_etimes,curr_etimes,wall_delta,expected",
    [
        # Continuous pid: Δetime exactly matches Δwall.
        (50.0, 110.0, 60.0, True),
        # 1-second slack from ps integer rounding -- still same pid.
        (50.0, 109.0, 60.0, True),
        # 2-second slack at the tolerance boundary -- still same pid.
        (50.0, 108.0, 60.0, True),
        # Concrete con/duct#399 case: pid 3323259 went 49->54 over 60s of
        # wall time. 5 << 58. Definitely a different physical process.
        (49.0, 54.0, 60.0, False),
        # etime regressed -- obvious reuse.
        (100.0, 5.0, 60.0, False),
        # Sub-quantum: same instant, no time elapsed for either etime or
        # wall -- treat as continuous (no reuse signal).
        (50.0, 50.0, 0.0, True),
    ],
)
def test_is_same_pid(
    prev_etimes: float,
    curr_etimes: float,
    wall_delta: float,
    expected: bool,
) -> None:
    assert is_same_pid(prev_etimes, curr_etimes, wall_delta) is expected


def test_is_same_pid_tolerance_kwarg() -> None:
    # 5s gap inside default tolerance? No (default is 2s).
    assert not is_same_pid(50.0, 105.0, 60.0)
    # ...but it is inside a 6s tolerance.
    assert is_same_pid(50.0, 105.0, 60.0, tolerance=6.0)



import argparse
from typing import Any
import unittest
from unittest.mock import MagicMock, mock_open, patch
from con_duct import cli, pprint_json
from con_duct._formatter import SummaryFormatter


class TestPPrint:

    @patch("con_duct.pprint_json.pprint")
    def test_pprint_json(self, mock_pprint: MagicMock, tmp_path: Any) -> None:
        json_file = tmp_path / "test.json"
        json_file.write_text('{"mock_key": "mock_value"}')

        args = argparse.Namespace(
            command="pp",
            file_path=str(json_file),
            func=pprint_json.pprint_json,
            log_level="INFO",
            humanize=False,
        )
        assert cli.execute(args) == 0
        mock_pprint.assert_called_once_with({"mock_key": "mock_value"})

    @patch("con_duct.pprint_json.pprint")
    def test_pprint_jsonl(self, mock_pprint: MagicMock, tmp_path: Any) -> None:
        jsonl_file = tmp_path / "test.jsonl"
        jsonl_file.write_text('{"line": 1}\n{"line": 2}\n')

        args = argparse.Namespace(
            command="pp",
            file_path=str(jsonl_file),
            func=pprint_json.pprint_json,
            log_level="INFO",
            humanize=False,
        )
        assert cli.execute(args) == 0
        mock_pprint.assert_called_once_with([{"line": 1}, {"line": 2}])

    def test_file_not_found(self) -> None:
        args = argparse.Namespace(
            command="pp",
            file_path="/nonexistent/path.json",
            func=pprint_json.pprint_json,
            log_level="INFO",
            humanize=False,
        )
        assert cli.execute(args) == 1

    @patch("con_duct.pprint_json.pprint")
    def test_pprint_invalid_json(self, mock_pprint: MagicMock, tmp_path: Any) -> None:
        json_file = tmp_path / "invalid.json"
        json_file.write_text('{"invalid": "json"')

        args = argparse.Namespace(
            command="pp",
            file_path=str(json_file),
            func=pprint_json.pprint_json,
            log_level="INFO",
            humanize=False,
        )
        assert cli.execute(args) == 1
        mock_pprint.assert_not_called()


class TestPPrintHumanization(unittest.TestCase):
    """Test humanization functionality in pprint_json module"""

    def test_apply_conversion_with_numbers(self) -> None:
        """Test _apply_conversion with numeric values"""

        formatter = SummaryFormatter()
        field_mapping = pprint_json.get_field_conversion_mapping()

        result = pprint_json._apply_conversion(
            "average_pcpu", 85.567, field_mapping, formatter
        )
        assert result == "85.57%"

        result = pprint_json._apply_conversion(
            "peak_rss", 1024000, field_mapping, formatter
        )
        assert result == "1.0 MB"

        result = pprint_json._apply_conversion(
            "memory_total", 1081801371648, field_mapping, formatter
        )
        assert result == "1.1 TB"

        result = pprint_json._apply_conversion(
            "wall_clock_time", 3661.5, field_mapping, formatter
        )
        assert result == "1h 1m 1.5s"

        result = pprint_json._apply_conversion(
            "start_time", 1625400000, field_mapping, formatter
        )
        assert "Jul" in result and "2021" in result

    def test_apply_conversion_with_non_numbers(self) -> None:
        """Test _apply_conversion with non-numeric values"""

        formatter = SummaryFormatter()
        field_mapping = pprint_json.get_field_conversion_mapping()

        # Test string values (should be returned unchanged)
        result = pprint_json._apply_conversion(
            "average_pcpu", "unknown", field_mapping, formatter
        )
        assert result == "unknown"

        # Test None values (should be returned unchanged)
        result = pprint_json._apply_conversion(
            "average_pcpu", None, field_mapping, formatter
        )
        assert result is None

    def test_apply_conversion_unmapped_fields(self) -> None:
        """Test _apply_conversion with unmapped field names"""

        formatter = SummaryFormatter()
        field_mapping = pprint_json.get_field_conversion_mapping()

        # Test unmapped field (should be returned unchanged)
        result = pprint_json._apply_conversion(
            "unknown_field", 42.0, field_mapping, formatter
        )
        assert result == 42.0

        # Test unmapped field (should be returned unchanged)
        result = pprint_json._apply_conversion(
            "unknown_field", "some_string", field_mapping, formatter
        )
        assert result == "some_string"

    def test_humanize_data_simple_dict(self) -> None:
        """Test humanize_data with a simple dictionary"""

        data = {
            "average_pcpu": 85.567,
            "average_rss": 1024000,
            "wall_clock_time": 150.75,
            "unknown_field": 42,
            "string_field": "test",
        }

        formatter = SummaryFormatter()
        result = pprint_json.humanize_data(data, formatter)

        assert result["average_pcpu"] == "85.57%"
        assert result["average_rss"] == "1.0 MB"
        assert result["wall_clock_time"] == "2m 30.8s"
        assert result["unknown_field"] == 42  # unchanged
        assert result["string_field"] == "test"  # unchanged

    def test_humanize_data_nested_structures(self) -> None:
        """Test humanize_data with nested dictionaries and lists"""

        data = {
            "process": {"peak_pcpu": 75.0, "average_vsz": 512000},
            "samples": [
                {"average_pcpu": 80.5, "peak_rss": 2048000},
                {"average_pcpu": 90.0, "peak_rss": 3072000},
            ],
        }

        formatter = SummaryFormatter()
        result = pprint_json.humanize_data(data, formatter)

        # Check nested dict
        assert result["process"]["peak_pcpu"] == "75.00%"
        assert result["process"]["average_vsz"] == "512.0 kB"

        # Check list of dicts
        assert result["samples"][0]["average_pcpu"] == "80.50%"
        assert result["samples"][0]["peak_rss"] == "2.0 MB"
        assert result["samples"][1]["average_pcpu"] == "90.00%"
        assert result["samples"][1]["peak_rss"] == "3.1 MB"

    @patch(
        "builtins.open",
        new_callable=mock_open,
        read_data='{"average_pcpu": 85.567, "average_rss": 1024000, "wall_clock_time": 150.75}',
    )
    @patch("con_duct.pprint_json.pprint")
    def test_pprint_json_with_humanize(
        self, mock_pprint: MagicMock, mock_open: MagicMock
    ) -> None:
        """Test pprint_json with humanize=True"""
        args = argparse.Namespace(
            command="pp",
            file_path="dummy.json",
            func=pprint_json.pprint_json,
            log_level="INFO",
            humanize=True,
        )

        assert cli.execute(args) == 0

        mock_open.assert_called_with("dummy.json", "r")

        # Verify that pprint was called with humanized data
        call_args = mock_pprint.call_args[0][0]
        assert call_args["average_pcpu"] == "85.57%"
        assert call_args["average_rss"] == "1.0 MB"
        assert call_args["wall_clock_time"] == "2m 30.8s"



from __future__ import annotations
import importlib
import sys
from typing import Generator
from unittest import mock
import pytest


@pytest.fixture
def _reload_sampling() -> Generator:
    """Ensure _sampling module is restored after tests that reload it."""
    import con_duct._sampling as mod

    yield
    if hasattr(sys, "tracebacklimit"):
        del sys.tracebacklimit
    importlib.reload(mod)


@mock.patch("platform.system", return_value="Windows")
@pytest.mark.usefixtures("_reload_sampling")
def test_unsupported_system_raises(_mock_system: mock.MagicMock) -> None:
    import con_duct._sampling as mod

    with pytest.raises(
        NotImplementedError, match="does not currently support.*Windows"
    ):
        importlib.reload(mod)



"""I/O utilities for con-duct."""

from __future__ import annotations
from collections.abc import Iterable
import os
import subprocess
import sys
import threading
import time
from typing import IO, Any, TextIO
from con_duct._models import LogPaths, Outputs


class TailPipe:
    """TailPipe simultaneously streams to an output stream (stdout or stderr) and a specified file."""

    TAIL_CYCLE_TIME = 0.01

    def __init__(self, file_path: str, buffer: IO[bytes]) -> None:
        self.file_path = file_path
        self.buffer = buffer
        self.stop_event: threading.Event | None = None
        self.infile: IO[bytes] | None = None
        self.thread: threading.Thread | None = None

    def start(self) -> None:
        self.stop_event = threading.Event()
        self.infile = open(self.file_path, "rb")
        self.thread = threading.Thread(target=self._tail, daemon=True)
        self.thread.start()

    def fileno(self) -> int:
        assert self.infile is not None
        return self.infile.fileno()

    def _catch_up(self) -> None:
        assert self.infile is not None
        data = self.infile.read()
        if data:
            self.buffer.write(data)
            self.buffer.flush()

    def _tail(self) -> None:
        assert self.stop_event is not None
        try:
            while not self.stop_event.is_set():
                self._catch_up()
                time.sleep(TailPipe.TAIL_CYCLE_TIME)
            # After stop event, collect and passthrough data one last time
            self._catch_up()
        except Exception:
            raise
        finally:
            self.buffer.flush()

    def close(self) -> None:
        assert self.stop_event is not None
        assert self.thread is not None
        assert self.infile is not None
        self.stop_event.set()
        self.thread.join()
        self.infile.close()


def prepare_outputs(
    capture_outputs: Outputs,
    outputs: Outputs,
    log_paths: LogPaths,
) -> tuple[TextIO | TailPipe | int | None, TextIO | TailPipe | int | None]:
    stdout: TextIO | TailPipe | int | None
    stderr: TextIO | TailPipe | int | None

    if capture_outputs.has_stdout():
        if outputs.has_stdout():
            stdout = TailPipe(log_paths.stdout, buffer=sys.stdout.buffer)
            stdout.start()
        else:
            stdout = open(log_paths.stdout, "w")
    elif outputs.has_stdout():
        stdout = None
    else:
        stdout = subprocess.DEVNULL

    if capture_outputs.has_stderr():
        if outputs.has_stderr():
            stderr = TailPipe(log_paths.stderr, buffer=sys.stderr.buffer)
            stderr.start()
        else:
            stderr = open(log_paths.stderr, "w")
    elif outputs.has_stderr():
        stderr = None
    else:
        stderr = subprocess.DEVNULL
    return stdout, stderr


def safe_close_files(file_list: Iterable[Any]) -> None:
    for f in file_list:
        try:
            f.close()
        except Exception:
            pass


def remove_files(log_paths: LogPaths, assert_empty: bool = False) -> None:
    for _, file_path in log_paths:
        if os.path.exists(file_path):
            if assert_empty:
                assert os.stat(file_path).st_size == 0
            os.remove(file_path)



import json
import os
from pathlib import Path
from utils import run_duct_command
from con_duct._constants import SUFFIXES
from con_duct.ls import LS_FIELD_CHOICES, _flatten_dict


def test_info_fields(temp_output_dir: str) -> None:
    """
    Generate the list of fields users can request when viewing info files.

    Fails when schema changes-- commit the new version and bump schema version
    """
    assert (
        run_duct_command(
            ["echo", "hello", "world"],
            sample_interval=4.0,
            report_interval=60.0,
            output_prefix=temp_output_dir,
        )
        == 0
    )
    os.remove(Path(temp_output_dir, SUFFIXES["stdout"]))
    os.remove(Path(temp_output_dir, SUFFIXES["stderr"]))
    os.remove(Path(temp_output_dir, SUFFIXES["usage"]))

    info_file = Path(temp_output_dir, SUFFIXES["info"])
    actual_info_schema = _flatten_dict(json.loads(info_file.read_text())).keys()
    os.remove(info_file)

    assert set(actual_info_schema) == set(LS_FIELD_CHOICES)



"""Summary formatter with custom conversions for con-duct output."""

from __future__ import annotations
from datetime import datetime
import logging
import string
from typing import Any

lgr = logging.getLogger("con-duct")


# Decimal (SI) byte units, single source of truth for byte humanization in
# duct. Used by SummaryFormatter.naturalsize for run-summary output and by
# the plot axis formatter for tick labels, so a "kB" means the same thing
# in both places. A future opt-in could add a 1024-base + IEC-suffix
# variant (KiB/MiB/GiB/...) and let callers pick.
FILESIZE_UNITS: list[tuple[str, int]] = [
    ("B", 1),
    ("kB", 1000**1),
    ("MB", 1000**2),
    ("GB", 1000**3),
    ("TB", 1000**4),
    ("PB", 1000**5),
    ("EB", 1000**6),
    ("ZB", 1000**7),
    ("YB", 1000**8),
]


class SummaryFormatter(string.Formatter):
    OK = "OK"
    NOK = "X"
    NONE = "-"
    BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(30, 38)
    RESET_SEQ = "\033[0m"
    COLOR_SEQ = "\033[1;%dm"

    def __init__(self, enable_colors: bool = False) -> None:
        self.enable_colors = enable_colors

    def naturalsize(
        self,
        value: float | str,
        format: str = "%.1f",  # noqa: A002
    ) -> str:
        """Format a number of bytes like a human readable decimal filesize (e.g. 10 kB).

        Examples:
            ```pycon
            >>> formatter = SummaryFormatter()
            >>> formatter.naturalsize(3000000)
            '3.0 MB'
            >>> formatter.naturalsize(3000, "%.3f")
            '3.000 kB'
            >>> formatter.naturalsize(10**28)
            '10000.0 YB'
            ```

        Args:
            value (int, float, str): Integer to convert.
            format (str): Custom formatter.

        Returns:
            str: Human readable representation of a filesize.
        """
        bytes_ = float(value)
        abs_bytes = abs(bytes_)

        if abs_bytes == 1:
            return "%d Byte" % bytes_

        if abs_bytes < 1000:
            return "%d Bytes" % bytes_

        # Pick the largest unit where the value is at least 1 of that unit.
        # FILESIZE_UNITS is ordered ascending; iterate to keep updating until
        # one further step would underflow.
        name, divisor = FILESIZE_UNITS[1]  # default to "kB" (loop entry value)
        for n, d in FILESIZE_UNITS[1:]:
            if abs_bytes / d >= 1:
                name, divisor = n, d
        return f"{format % (bytes_ / divisor)} {name}"

    def color_word(self, s: str, color: int) -> str:
        """Color `s` with `color`.

        Parameters
        ----------
        s : string
        color : int
            Code for color. If the value evaluates to false, the string will not be
            colored.
        enable_colors: boolean, optional

        Returns
        -------
        str
        """
        if color and self.enable_colors:
            return "%s%s%s" % (self.COLOR_SEQ % color, s, self.RESET_SEQ)
        return s

    def _format_duration(self, value: float) -> str:
        """Format a duration in seconds to human-readable format."""
        if value >= 3600:  # >= 1 hour
            hours = int(value // 3600)
            minutes = int((value % 3600) // 60)
            seconds = value % 60
            return f"{hours}h {minutes}m {seconds:.1f}s"
        elif value >= 60:  # >= 1 minute
            minutes = int(value // 60)
            seconds = value % 60
            return f"{minutes}m {seconds:.1f}s"
        else:
            return f"{value:.2f}s"

    def convert_field(self, value: str | None, conversion: str | None) -> Any:
        if conversion == "S":  # Human size
            if value is not None:
                return self.color_word(self.naturalsize(value), self.GREEN)
            else:
                return self.color_word(self.NONE, self.RED)
        elif conversion == "E":  # colored non-zero is bad
            return self.color_word(
                value if value is not None else self.NONE,
                self.RED if value or value is None else self.GREEN,
            )
        elif conversion == "X":  # colored truthy
            col = self.GREEN if value else self.RED
            return self.color_word(value if value is not None else self.NONE, col)
        elif conversion == "N":  # colored Red - if None
            if value is None:
                return self.color_word(self.NONE, self.RED)
            else:
                return self.color_word(value, self.GREEN)
        elif conversion == "P":  # Percentage
            if value is not None:
                return f"{float(value):.2f}%"
            else:
                return self.color_word(self.NONE, self.RED)
        elif conversion == "T":  # Time duration
            if value is not None:
                return self._format_duration(float(value))
            else:
                return self.color_word(self.NONE, self.RED)
        elif conversion == "D":  # DateTime from timestamp
            if value is not None:
                try:
                    dt = datetime.fromtimestamp(float(value))
                    return dt.strftime("%b %d, %Y %I:%M %p")
                except (ValueError, OSError):
                    return str(value)
            else:
                return self.color_word(self.NONE, self.RED)

        return super().convert_field(value, conversion)

    def format_field(self, value: Any, format_spec: str) -> Any:
        # TODO: move all the "coloring" into formatting, so we could correctly indent
        # given the format and only then color it up
        # print "> %r, %r" % (value, format_spec)
        if value is None:
            # TODO: could still use our formatter and make it red or smth like that
            return self.NONE
        # if it is a composite :format!conversion, we need to split it
        if "!" in format_spec and format_spec.index("!") > 1:
            format_spec, conversion = format_spec.split("!", 1)
        else:
            conversion = None
        try:
            value_ = super().format_field(value, format_spec)
        except ValueError as exc:
            lgr.warning(
                "Falling back to `str` formatting for %r due to exception: %s",
                value,
                exc,
            )
            return str(value)
        if conversion:
            return self.convert_field(value_, conversion)
        return value_



"""Tests for plot command."""

import argparse
import json
import os
from pathlib import Path
from typing import Any, List, Tuple
from unittest.mock import MagicMock, Mock, mock_open, patch
import pytest

pytest.importorskip("matplotlib")
from con_duct import cli, plot  # noqa: E402
from con_duct._formatter import FILESIZE_UNITS  # noqa: E402


@pytest.mark.parametrize(
    "min_ratio,span_seconds,expected_unit",
    [
        # min_ratio=-1: always use base unit
        (-1, 2, "s"),  # Small value
        (-1, 3700, "s"),  # More than 1 hour - still base unit
        (-1, 3 * 60 * 60 * 24, "s"),  # 3 days - still base unit
        # min_ratio=1.5: switch units more aggressively
        (1.5, 90, "min"),  # 1.5 minutes meets threshold
        (1.5, 90 * 60, "h"),  # 1.5 hours meets threshold
        (1.5, 36 * 60 * 60, "d"),  # 1.5 days meets threshold
        # min_ratio=3.0: standard threshold
        (3.0, 2, "s"),  # 2 seconds - stays in base unit
        (3.0, 3 * 60, "min"),  # 3 minutes - meets min_ratio for minutes
        (3.0, 3 * 60 * 60, "h"),  # 3 hours - meets min_ratio for hours
        (3.0, 3 * 60 * 60 * 24, "d"),  # 3 days - meets min_ratio for days
        # min_ratio=5.0: more conservative switching
        (5.0, 4 * 60, "s"),  # 4 minutes - doesn't meet threshold, stays seconds
        (5.0, 5 * 60, "min"),  # 5 minutes - meets threshold
        (5.0, 4 * 60 * 60, "min"),  # 4 hours - doesn't meet hour threshold
        (5.0, 5 * 60 * 60, "h"),  # 5 hours - meets hour threshold
    ],
)
def test_pick_unit_with_varying_ratios(
    min_ratio: float, span_seconds: float, expected_unit: str
) -> None:
    """Test pick_unit selects appropriate unit based on min_ratio."""
    formatter: Any = plot.HumanizedAxisFormatter(
        min_ratio=min_ratio, units=plot._TIME_UNITS
    )
    unit_name, _ = formatter.pick_unit(span_seconds)
    assert unit_name == expected_unit


@pytest.mark.parametrize(
    "units,axis_range,value,expected",
    [
        # Time formatting tests
        (plot._TIME_UNITS, (0, 30), 15, "15.0s"),
        (plot._TIME_UNITS, (0, 300), 2.3 * 60, "2.3min"),
        (plot._TIME_UNITS, (0, 11000), 7.8 * 60 * 60, "7.8h"),
        (plot._TIME_UNITS, (0, 260000), 3.2 * 60 * 60 * 24, "3.2d"),
        # Memory formatting tests (base 1000, kB/MB/GB/TB/PB).
        (FILESIZE_UNITS, (0, 5 * 1000), 2.6 * 1000, "2.6kB"),
        (FILESIZE_UNITS, (0, 4 * 1000**2), 1.5 * (1000**2), "1.5MB"),
        (FILESIZE_UNITS, (0, 3 * 1000**3), 8.3 * 1000**3, "8.3GB"),
        (FILESIZE_UNITS, (0, 3 * 1000**4), 1.3 * 1000**4, "1.3TB"),
        (FILESIZE_UNITS, (0, 3.1 * 1000**5), 6.5 * 1000**5, "6.5PB"),
    ],
)
def test_formatter_output(
    units: List[Tuple[str, float]],
    axis_range: Tuple[float, float],
    value: float,
    expected: str,
) -> None:
    """Test HumanizedAxisFormatter formats values correctly for time and memory units."""
    formatter: Any = plot.HumanizedAxisFormatter(min_ratio=3.0, units=units)
    formatter.axis = Mock()
    formatter.axis.get_view_interval.return_value = axis_range
    result = formatter(value, 0)
    assert result == expected


class TestPlotMatplotlib:

    @patch("matplotlib.pyplot.savefig")
    def test_matplotlib_plot_sanity(self, mock_plot_save: MagicMock) -> None:
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        assert cli.execute(args) == 0
        mock_plot_save.assert_called_once_with("outfile.png")

    @patch("matplotlib.pyplot.savefig")
    @patch("matplotlib.use")
    def test_matplotlib_plot_uses_agg_backend_with_output(
        self, mock_use: MagicMock, mock_plot_save: MagicMock
    ) -> None:
        """Test that Agg backend is used when --output is specified."""
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        assert cli.execute(args) == 0
        mock_use.assert_called_once_with("Agg")
        mock_plot_save.assert_called_once_with("outfile.png")

    @patch("matplotlib.pyplot.savefig")
    def test_matplotlib_plot_file_not_found(self, mock_plot_save: MagicMock) -> None:
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage_not_to_be_found.json",
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
        )
        assert cli.execute(args) == 1
        mock_plot_save.assert_not_called()

    @patch("matplotlib.pyplot.savefig")
    @patch("builtins.open", new_callable=mock_open, read_data='{"invalid": "json"')
    def test_matplotlib_plot_invalid_json(
        self, _mock_open: MagicMock, mock_plot_save: MagicMock
    ) -> None:
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
        )
        assert cli.execute(args) == 1
        mock_plot_save.assert_not_called()

    @patch("matplotlib.pyplot.savefig")
    def test_matplotlib_plot_info_json(self, mock_plot_save: MagicMock) -> None:
        """When user passes info.json, usage.json is retrieved and used"""
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/info.json",
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        assert cli.execute(args) == 0
        mock_plot_save.assert_called_once_with("outfile.png")

    @patch("matplotlib.pyplot.savefig")
    def test_matplotlib_plot_info_json_absolute_path(
        self, mock_plot_save: MagicMock, monkeypatch: Any, tmp_path: Any
    ) -> None:
        """Test that absolute path to info.json correctly resolves usage.json path
        when cwd is not the original execution wd.
        """
        abs_info_path = os.path.abspath("test/data/mriqc-example/info.json")

        # change into a pytest-managed temporary directory
        monkeypatch.chdir(tmp_path)

        args = argparse.Namespace(
            command="plot",
            file_path=abs_info_path,
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        assert cli.execute(args) == 0
        mock_plot_save.assert_called_once_with("outfile.png")

    @patch("matplotlib.pyplot.savefig")
    @patch(
        "builtins.open", new_callable=mock_open, read_data='{"missing": "timestamp"}'
    )
    def test_matplotlib_plot_malformed_usage_file(
        self, _mock_open: MagicMock, mock_plot_save: MagicMock
    ) -> None:
        """Test that malformed usage.json files are handled gracefully"""
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/malformed_usage.json",
            output="outfile.png",
            func=plot.matplotlib_plot,
            log_level="INFO",
        )
        assert cli.execute(args) == 1
        mock_plot_save.assert_not_called()

    @patch(
        "matplotlib.get_backend",
        side_effect=AttributeError("get_backend not available"),
    )
    @patch.dict("matplotlib.rcParams", {"backend": "Agg"})
    def test_matplotlib_plot_non_interactive_backend(
        self,
        _mock_get_backend: MagicMock,
    ) -> None:
        """Test that plotting without output in non-interactive backend returns error."""
        import matplotlib.backends

        if not hasattr(matplotlib.backends, "backend_registry"):
            pytest.skip("requires backend_registry (matplotlib >= 3.9)")

        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output=None,  # No output file specified
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        result = cli.execute(args)
        assert result == 1

    @patch("matplotlib.get_backend", return_value="Agg")
    def test_matplotlib_plot_non_interactive_backend_with_get_backend(
        self,
        _mock_get_backend: MagicMock,
    ) -> None:
        """Test that plotting without output in non-interactive backend returns error using get_backend."""
        import matplotlib.backends

        if not hasattr(matplotlib.backends, "backend_registry"):
            pytest.skip("requires backend_registry (matplotlib >= 3.9)")

        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output=None,  # No output file specified
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        result = cli.execute(args)
        assert result == 1

    @patch("matplotlib.pyplot.show")
    @patch("matplotlib.get_backend", return_value="tkagg")
    def test_matplotlib_plot_interactive_backend_with_get_backend(
        self,
        _mock_get_backend: MagicMock,
        mock_show: MagicMock,
    ) -> None:
        """Test that plotting without output in interactive backend calls plt.show() successfully."""

        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output=None,  # No output file specified
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        result = cli.execute(args)
        assert result == 0
        mock_show.assert_called_once()

    @patch(
        "builtins.__import__", side_effect=ImportError("No module named 'matplotlib'")
    )
    def test_matplotlib_plot_missing_dependency(self, _mock_import: MagicMock) -> None:
        """Test that plotting with missing matplotlib shows helpful error."""
        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output=None,
            func=plot.matplotlib_plot,
            log_level="INFO",
        )

        result = cli.execute(args)
        assert result == 1

    @patch("matplotlib.pyplot.show")
    def test_matplotlib_plot_no_backend_registry(
        self,
        mock_show: MagicMock,
        monkeypatch: Any,
        caplog: Any,
    ) -> None:
        """Test fallback when backend_registry unavailable (matplotlib < 3.9)."""
        import matplotlib.backends

        # Only delete if present (already absent on matplotlib < 3.9)
        if hasattr(matplotlib.backends, "backend_registry"):
            monkeypatch.delattr(matplotlib.backends, "backend_registry")

        args = argparse.Namespace(
            command="plot",
            file_path="test/data/mriqc-example/usage.json",
            output=None,
            func=plot.matplotlib_plot,
            log_level="INFO",
            min_ratio=3.0,
            cpu="ps-pcpu",
        )
        result = cli.execute(args)
        assert result == 0
        mock_show.assert_called_once()
        assert "matplotlib < 3.9" in caplog.text


class TestBuildPidSeriesCpuMode:
    """``cpu`` series content varies by ``cpu_mode``."""

    @staticmethod
    def _record(ts: str, pid: str, pcpu: float, etime: str) -> dict:
        return {
            "timestamp": ts,
            "processes": {
                pid: {
                    "pcpu": pcpu,
                    "pmem": 0.0,
                    "rss": 0,
                    "etime": etime,
                    "timestamp": ts,
                    "cmd": "x",
                }
            },
        }

    def test_ps_pcpu_takes_raw_values(self) -> None:
        data = [
            self._record("2026-05-06T00:00:00", "1", 1.5, "00:01"),
            self._record("2026-05-06T00:01:00", "1", 2.5, "01:01"),
            self._record("2026-05-06T00:02:00", "1", 3.5, "02:01"),
        ]
        series = plot._build_pid_series(data, cpu_mode=plot.CPU_MODE_PS_PCPU)
        assert series["1"]["cpu"] == [1.5, 2.5, 3.5]

    def test_ps_pcpu_includes_etime_zero_record(self) -> None:
        # In timepoint mode etime=="00:00" is dropped before delta math; in
        # raw mode every record contributes its pcpu unchanged.
        data = [
            self._record("2026-05-06T00:00:00", "1", 0.0, "00:00"),
            self._record("2026-05-06T00:01:00", "1", 5.0, "01:00"),
        ]
        series = plot._build_pid_series(data, cpu_mode=plot.CPU_MODE_PS_PCPU)
        assert series["1"]["cpu"] == [0.0, 5.0]

    def test_ps_cpu_timepoint_drops_first_and_etime_zero(self) -> None:
        # Baseline: timepoint mode unchanged by the new flag -- first record
        # is None (no prior baseline), etime=="00:00" record is None.
        data = [
            self._record("2026-05-06T00:00:00", "1", 10.0, "01:00"),
            self._record("2026-05-06T00:01:00", "1", 0.0, "00:00"),
            self._record("2026-05-06T00:02:00", "1", 50.0, "03:00"),
        ]
        series = plot._build_pid_series(data, cpu_mode=plot.CPU_MODE_PS_CPU_TIMEPOINT)
        cpu = series["1"]["cpu"]
        assert cpu[0] is None
        assert cpu[1] is None

    def test_default_is_ps_pcpu(self) -> None:
        data = [self._record("2026-05-06T00:00:00", "1", 7.0, "00:30")]
        series = plot._build_pid_series(data)
        assert series["1"]["cpu"] == [7.0]


class TestLoadHostMemoryTotal:
    """Best-effort host memory_total lookup for the rss legend label."""

    def _write_info(self, path: Path, data: dict) -> None:
        path.write_text(json.dumps(data))

    def test_info_json_input(self, tmp_path: Path) -> None:
        info = tmp_path / "run_info.json"
        self._write_info(info, {"system": {"memory_total": 12345}})
        assert plot._load_host_memory_total(info) == 12345

    def test_usage_jsonl_with_sibling(self, tmp_path: Path) -> None:
        info = tmp_path / "run_info.json"
        usage = tmp_path / "run_usage.jsonl"
        self._write_info(info, {"system": {"memory_total": 67890}})
        usage.write_text("")
        assert plot._load_host_memory_total(usage) == 67890

    def test_usage_legacy_with_sibling(self, tmp_path: Path) -> None:
        info = tmp_path / "run_info.json"
        usage = tmp_path / "run_usage.json"
        self._write_info(info, {"system": {"memory_total": 42}})
        usage.write_text("")
        assert plot._load_host_memory_total(usage) == 42

    def test_no_sibling_returns_none(self, tmp_path: Path) -> None:
        usage = tmp_path / "run_usage.jsonl"
        usage.write_text("")
        assert plot._load_host_memory_total(usage) is None

    def test_missing_key_returns_none(self, tmp_path: Path) -> None:
        info = tmp_path / "run_info.json"
        self._write_info(info, {"system": {}})
        assert plot._load_host_memory_total(info) is None

    def test_invalid_json_returns_none(self, tmp_path: Path) -> None:
        info = tmp_path / "run_info.json"
        info.write_text("{not json")
        assert plot._load_host_memory_total(info) is None

    def test_unparseable_filename_returns_none(self, tmp_path: Path) -> None:
        weird = tmp_path / "weird.txt"
        weird.write_text("")
        assert plot._load_host_memory_total(weird) is None



"""Utility functions for con-duct."""

from typing import Any


def assert_num(*values: Any) -> None:
    for value in values:
        assert isinstance(value, (float, int))


# TODO: consider asking ps for `etimes` (seconds) directly via
# `-o etimes` instead of parsing `etime`. Even if we switch, this
# parser is worth keeping for backwards compatibility with existing
# usage.jsonl logs that persist `etime` as a string.
def etime_to_etimes(etime: str) -> float:
    """Parse a ps ``etime`` string into seconds.

    ps's ``etime`` format is ``[[DD-]HH:]MM:SS``: ``MM:SS`` always,
    with ``HH:`` prepended after one hour and ``DD-`` prepended after
    one day.

    :param etime: elapsed-time string from ``ps -o etime``.
    :returns: elapsed time in seconds.
    :raises ValueError: if ``etime`` does not match the expected shape.
    """
    if "-" in etime:
        days_str, rest = etime.split("-", 1)
        days = int(days_str)
    else:
        days = 0
        rest = etime
    parts = rest.split(":")
    if len(parts) == 2:
        hours, minutes, seconds = 0, int(parts[0]), int(parts[1])
    elif len(parts) == 3:
        hours, minutes, seconds = int(parts[0]), int(parts[1]), int(parts[2])
    else:
        raise ValueError(f"Unparsable etime: {etime!r}")
    return float(days * 86400 + hours * 3600 + minutes * 60 + seconds)


def is_same_pid(
    prev_etimes: float,
    curr_etimes: float,
    wall_delta: float,
    *,
    tolerance: float = 2.0,
) -> bool:
    """Decide whether two consecutive observations of the same pid number
    correspond to the same physical process (vs kernel pid reuse).

    A continuously-existing pid's ``etime`` grows by exactly the
    wall-clock time between two samples (etime is wall-since-fork,
    not cputime). ps reports ``etime`` at integer-second resolution,
    so the only slack is rounding plus small clock drift.

    A stricter ``etime_2 < etime_1`` check would miss reuses where
    the new pid's etime crept above the old pid's last reading by
    less than the sample interval. Comparing ``Δetime`` against
    ``Δwall`` catches those too.

    :param prev_etimes: elapsed seconds at the earlier sample.
    :param curr_etimes: elapsed seconds at the later sample.
    :param wall_delta: wall-clock seconds between the two samples.
    :param tolerance: slack (seconds) absorbed by ps's integer-second
        ``etime`` plus clock drift. 2 seconds is plenty in practice.
    :returns: ``True`` iff ``Δetime`` is close enough to ``Δwall``
        to be the same continuous process.
    """
    return curr_etimes - prev_etimes >= wall_delta - tolerance


def pdcpu_from_pcpu(
    prev_pcpu: float,
    prev_etimes: float,
    curr_pcpu: float,
    curr_etimes: float,
) -> float | None:
    """Delta-corrected %CPU between two ps samples of the same pid.

    Inverts the procps identity ``pcpu = cputime / etime * 100`` to
    recover cputime at each sample, takes the cputime delta, and
    divides by the elapsed interval. The ``/100`` and ``*100``
    cancel, so the result is in the same units as ``pcpu``.

    Linux-only: assumes ``pcpu`` is the cumulative ``cputime/etime``
    ratio. Invalid on Darwin (decayed EWMA).

    Identity is the caller's responsibility: invoke ``is_same_pid``
    first. If the inputs satisfy that precondition, ``Δetime`` is
    positive and the math is well-defined.

    :param prev_pcpu: %CPU from the earlier sample.
    :param prev_etimes: elapsed seconds at the earlier sample.
    :param curr_pcpu: %CPU from the later sample.
    :param curr_etimes: elapsed seconds at the later sample.
    :returns: delta-corrected %CPU over the interval, or ``None``
        in two "no measurement" cases:

        - ``Δetime <= 0`` -- defensive guard for callers that
          skipped ``is_same_pid``; covers sub-quantum and obvious
          pid-reuse.
        - Computed ``pdcpu < 0`` -- aggregation-timing artifact.
          When duct's per-pid ``pcpu`` is max-across-samples while
          ``etime`` is from the last sample, a spike-then-idle
          pattern earlier in the run inflates ``prev_pcpu *
          prev_etimes`` enough that the cputime "delta" goes
          negative. The pid is the same; the math is just noisy.
          A small minus dip is honestly null, not zero.
    """
    interval = curr_etimes - prev_etimes
    if interval <= 0:
        return None
    pdcpu = (curr_pcpu * curr_etimes - prev_pcpu * prev_etimes) / interval
    if pdcpu < 0:
        return None
    return pdcpu


def parse_version(version_str: str) -> tuple[int, int, int]:
    x_y_z = version_str.split(".")
    if len(x_y_z) != 3:
        raise ValueError(
            f"Invalid version format: {version_str}. Expected 'x.y.z' format."
        )

    x, y, z = map(int, x_y_z)  # Unpacking forces exactly 3 elements
    return (x, y, z)



"""Process tracking and reporting for con-duct."""

from __future__ import annotations
from dataclasses import asdict
from importlib.metadata import version
import json
import logging
import math
import os
import shutil
import socket
import subprocess
import threading
import time
from typing import Any, Optional, TextIO
from con_duct._constants import ENV_PREFIXES, __schema_version__
from con_duct._formatter import SummaryFormatter
from con_duct._models import LogPaths, Sample, SystemInfo
from con_duct._output import safe_close_files
from con_duct._sampling import _get_sample

__version__ = version("con-duct")

lgr = logging.getLogger("con-duct")


class Report:
    """Top level report"""

    def __init__(
        self,
        command: str,
        arguments: list[str],
        log_paths: LogPaths,
        summary_format: str,
        working_directory: str,
        colors: bool = False,
        clobber: bool = False,
        process: subprocess.Popen | None = None,
        message: str = "",
    ) -> None:
        self._command = command
        self.arguments = arguments
        self.log_paths = log_paths
        self.summary_format: str = summary_format
        self.clobber = clobber
        self.colors = colors
        self.message = message
        # Defaults to be set later
        self.start_time: float | None = None
        self.process = process
        self.session_id: int | None = None
        self.gpus: list[dict[str, str]] | None = None
        self.env: dict[str, str] | None = None
        self.number = 1
        self.system_info: SystemInfo | None = None
        self.full_run_stats = Sample()
        self.current_sample: Optional[Sample] = None
        self.end_time: float | None = None
        self.run_time_seconds: str | None = None
        self.usage_file: TextIO | None = None
        self.working_directory: str = working_directory

    def __del__(self) -> None:
        safe_close_files([self.usage_file])

    @property
    def command(self) -> str:
        return " ".join([self._command] + self.arguments)

    @property
    def elapsed_time(self) -> float:
        assert self.start_time is not None
        return time.time() - self.start_time

    @property
    def wall_clock_time(self) -> Optional[float]:
        if self.start_time is None:
            return math.nan
        if self.end_time is None:
            # if no end_time -- must be still ongoing
            # Cannot happen ATM but could in "library mode" later
            return time.time() - self.start_time
        # we reached the end
        return self.end_time - self.start_time

    def collect_environment(self) -> None:
        self.env = {k: v for k, v in os.environ.items() if k.startswith(ENV_PREFIXES)}

    def get_system_info(self) -> None:
        """Gathers system information related to CPU, GPU, memory, and environment variables."""
        self.system_info = SystemInfo(
            cpu_total=os.sysconf("SC_NPROCESSORS_CONF"),
            memory_total=os.sysconf("SC_PAGESIZE") * os.sysconf("SC_PHYS_PAGES"),
            hostname=socket.gethostname(),
            uid=os.getuid(),
            user=os.environ.get("USER"),
        )
        # GPU information
        if shutil.which("nvidia-smi") is not None:
            lgr.debug("Checking NVIDIA GPU using nvidia-smi")
            try:
                out = subprocess.check_output(
                    [
                        "nvidia-smi",
                        "--query-gpu=index,name,pci.bus_id,driver_version,memory.total,compute_mode",
                        "--format=csv",
                    ]
                )
            except subprocess.CalledProcessError as e:
                lgr.warning("Error collecting gpu information: %s", str(e))
                self.gpus = None
                return
            try:
                decoded = out.decode("utf-8")
                lines = decoded.strip().split("\n")
                _ = lines.pop(0)  # header
                self.gpus = []
                for line in lines:
                    cols = line.split(", ")
                    self.gpus.append(
                        {
                            "index": cols[0],
                            "name": cols[1],
                            "bus_id": cols[2],
                            "driver_version": cols[3],
                            "memory.total": cols[4],
                            "compute_mode": cols[5],
                        }
                    )
            except Exception as e:
                lgr.warning("Error parsing gpu information: %s", str(e))
                self.gpus = None

    def collect_sample(self) -> Optional[Sample]:
        assert self.session_id is not None
        try:
            sample = _get_sample(self.session_id)
            return sample
        except subprocess.CalledProcessError as exc:  # when session_id has no processes
            lgr.debug("Error collecting sample: %s", str(exc))
            return None

    def update_from_sample(self, sample: Sample) -> None:
        self.full_run_stats = self.full_run_stats.aggregate(sample)
        if self.current_sample is None:
            self.current_sample = Sample().aggregate(sample)
        else:
            assert self.current_sample.averages is not None
            self.current_sample = self.current_sample.aggregate(sample)
        assert self.current_sample is not None

    def write_subreport(self) -> None:
        assert self.current_sample is not None
        if self.usage_file is None:
            self.usage_file = open(self.log_paths.usage, "w")
        self.usage_file.write(json.dumps(self.current_sample.for_json()) + "\n")
        self.usage_file.flush()  # Force flush immediately

    @property
    def execution_summary(self) -> dict[str, Any]:
        # killed by a signal
        # https://pubs.opengroup.org/onlinepubs/9799919799/utilities/V3_chap02.html#tag_19_08_02
        if self.process and self.process.returncode < 0:
            self.process.returncode = 128 + abs(self.process.returncode)
        # prepare the base, but enrich if we did get process running
        return {
            "exit_code": self.process.returncode if self.process else None,
            "command": self.command,
            "logs_prefix": self.log_paths.prefix if self.log_paths else "",
            "wall_clock_time": self.wall_clock_time,
            "peak_rss": self.full_run_stats.total_rss,
            "average_rss": self.full_run_stats.averages.rss,
            "peak_vsz": self.full_run_stats.total_vsz,
            "average_vsz": self.full_run_stats.averages.vsz,
            "peak_pmem": self.full_run_stats.total_pmem,
            "average_pmem": self.full_run_stats.averages.pmem,
            "peak_pcpu": self.full_run_stats.total_pcpu,
            "average_pcpu": self.full_run_stats.averages.pcpu,
            "num_samples": self.full_run_stats.averages.num_samples,
            "num_reports": self.number,
            "start_time": self.start_time,
            "end_time": self.end_time,
            "working_directory": self.working_directory,
        }

    def dump_json(self) -> str:
        return json.dumps(
            {
                "command": self.command,
                "system": (
                    None if self.system_info is None else asdict(self.system_info)
                ),
                "env": self.env,
                "gpu": self.gpus,
                "duct_version": __version__,
                "schema_version": __schema_version__,
                "execution_summary": self.execution_summary,
                "output_paths": asdict(self.log_paths),
                "working_directory": self.working_directory,
                "message": self.message,
            }
        )

    @property
    def execution_summary_formatted(self) -> str:
        formatter = SummaryFormatter(enable_colors=self.colors)
        return formatter.format(self.summary_format, **self.execution_summary)


def monitor_process(
    report: Report,
    process: subprocess.Popen,
    report_interval: float,
    sample_interval: float,
    stop_event: threading.Event,
) -> None:
    lgr.debug(
        "Starting monitoring of the process %s on sample interval %f for report interval %f",
        process,
        sample_interval,
        report_interval,
    )
    while True:
        if process.poll() is not None:
            lgr.debug(
                "Breaking out of the monitor since the passthrough command has finished"
            )
            break
        sample = report.collect_sample()
        # Report averages should be updated prior to sample aggregation
        if (
            sample is None
        ):  # passthrough has probably finished before sample could be collected
            if process.poll() is not None:
                lgr.debug(
                    "Breaking out of the monitor since the passthrough command has finished "
                    "before we could collect sample"
                )
                break
            # process is still running, but we could not collect sample
            continue
        report.update_from_sample(sample)
        if (
            report.start_time
            and report.elapsed_time >= (report.number - 1) * report_interval
        ):
            report.write_subreport()
            report.current_sample = None
            report.number += 1
        if stop_event.wait(timeout=sample_interval):
            lgr.debug("Breaking out because stop event was set")
            break



"""Data models and enums for con-duct."""

from __future__ import annotations
from collections import Counter
from collections.abc import Iterator
from dataclasses import asdict, dataclass, field
from datetime import datetime
from enum import Enum
import logging
import os
import re
import string
from typing import Any, Optional
from con_duct._constants import SUFFIXES
from con_duct._utils import assert_num

lgr = logging.getLogger("con-duct")


class Outputs(str, Enum):
    ALL = "all"
    NONE = "none"
    STDOUT = "stdout"
    STDERR = "stderr"

    def __str__(self) -> str:
        return self.value

    def has_stdout(self) -> bool:
        return self is Outputs.ALL or self is Outputs.STDOUT

    def has_stderr(self) -> bool:
        return self is Outputs.ALL or self is Outputs.STDERR


class RecordTypes(str, Enum):
    ALL = "all"
    SYSTEM_SUMMARY = "system-summary"
    PROCESSES_SAMPLES = "processes-samples"

    def __str__(self) -> str:
        return self.value

    def has_system_summary(self) -> bool:
        return self is RecordTypes.ALL or self is RecordTypes.SYSTEM_SUMMARY

    def has_processes_samples(self) -> bool:
        return self is RecordTypes.ALL or self is RecordTypes.PROCESSES_SAMPLES


class SessionMode(str, Enum):
    NEW_SESSION = "new-session"
    CURRENT_SESSION = "current-session"

    def __str__(self) -> str:
        return self.value


@dataclass
class SystemInfo:
    cpu_total: int
    memory_total: int
    hostname: str | None
    uid: int
    user: str | None


@dataclass
class ProcessStats:
    pcpu: float  # %CPU
    pmem: float  # %MEM
    rss: int  # Memory Resident Set Size in Bytes
    vsz: int  # Virtual Memory size in Bytes
    timestamp: str
    etime: str
    stat: Counter
    cmd: str

    def aggregate(self, other: ProcessStats) -> ProcessStats:
        cmd = self.cmd
        if self.cmd != other.cmd:
            lgr.debug(
                f"cmd has changed. Previous measurement was {self.cmd}, now {other.cmd}."
            )
            # Brackets indicate that the kernel has substituted an abbreviation.
            surrounded_by_brackets = r"^\[.+\]"
            if re.search(surrounded_by_brackets, self.cmd):
                lgr.debug(f"using {other.cmd}.")
                cmd = other.cmd
            lgr.debug(f"using {self.cmd}.")

        new_counter: Counter = Counter()
        new_counter.update(self.stat)
        new_counter.update(other.stat)
        return ProcessStats(
            pcpu=max(self.pcpu, other.pcpu),
            pmem=max(self.pmem, other.pmem),
            rss=max(self.rss, other.rss),
            vsz=max(self.vsz, other.vsz),
            timestamp=max(self.timestamp, other.timestamp),
            etime=other.etime,  # For the aggregate always take the latest
            stat=new_counter,
            cmd=cmd,
        )

    def for_json(self) -> dict:
        ret = asdict(self)
        ret["stat"] = dict(self.stat)
        return ret

    def __post_init__(self) -> None:
        self._validate()

    def _validate(self) -> None:
        assert_num(self.pcpu, self.pmem, self.rss, self.vsz)


@dataclass
class LogPaths:
    stdout: str
    stderr: str
    usage: str
    info: str
    prefix: str

    def __iter__(self) -> Iterator[tuple[str, str]]:
        for name, path in asdict(self).items():
            if name != "prefix":
                yield name, path

    @classmethod
    def create(cls, output_prefix: str, pid: None | int = None) -> LogPaths:
        if any(
            field_name == "datetime_filesafe"
            for _, field_name, _, _ in string.Formatter().parse(output_prefix)
        ):
            lgr.warning(
                "{datetime_filesafe} in --output-prefix is deprecated; use {datetime} instead"
            )
        datetime_filesafe = datetime.now().strftime("%Y.%m.%dT%H.%M.%S")
        formatted_prefix = output_prefix.format(
            pid=pid,
            datetime=datetime_filesafe,
            # Use of the `datetime_filesafe` format field is deprecated.
            # The setting of it here is to provide for backwards compatibility
            # It should be removed eventually
            datetime_filesafe=datetime_filesafe,
        )
        return cls(
            stdout=f"{formatted_prefix}{SUFFIXES['stdout']}",
            stderr=f"{formatted_prefix}{SUFFIXES['stderr']}",
            usage=f"{formatted_prefix}{SUFFIXES['usage']}",
            info=f"{formatted_prefix}{SUFFIXES['info']}",
            prefix=formatted_prefix,
        )

    def prepare_paths(self, clobber: bool, capture_outputs: Outputs) -> None:
        conflicts = [path for _name, path in self if os.path.exists(path)]
        if conflicts and not clobber:
            raise FileExistsError(
                "Conflicting files:\n"
                + "\n".join(f"- {path}" for path in conflicts)
                + "\nUse --clobber to overwrite conflicting files."
            )

        if self.prefix.endswith(os.sep):  # If it ends in "/" (for linux) treat as a dir
            os.makedirs(self.prefix, exist_ok=True)
        else:
            # Path does not end with a separator, treat the last part as a filename
            directory = os.path.dirname(self.prefix)
            if directory:
                os.makedirs(directory, exist_ok=True)
        for name, path in self:
            if name == SUFFIXES["stdout"] and not capture_outputs.has_stdout():
                continue
            elif name == SUFFIXES["stderr"] and not capture_outputs.has_stderr():
                continue
            # TODO: AVOID PRECREATION -- would interfere e.g. with git-annex
            # assistant monitoring new files to be created and committing
            # as soon as they are closed
            open(path, "w").close()


@dataclass
class Averages:
    rss: Optional[float] = None
    vsz: Optional[float] = None
    pmem: Optional[float] = None
    pcpu: Optional[float] = None
    num_samples: int = 0

    def update(self: Averages, other: Sample) -> None:
        assert_num(other.total_rss, other.total_vsz, other.total_pmem, other.total_pcpu)
        if not self.num_samples:
            self.num_samples += 1
            self.rss = other.total_rss
            self.vsz = other.total_vsz
            self.pmem = other.total_pmem
            self.pcpu = other.total_pcpu
        else:
            assert self.rss is not None
            assert self.vsz is not None
            assert self.pmem is not None
            assert self.pcpu is not None
            assert other.total_rss is not None
            assert other.total_vsz is not None
            assert other.total_pmem is not None
            assert other.total_pcpu is not None
            self.num_samples += 1
            self.rss += (other.total_rss - self.rss) / self.num_samples
            self.vsz += (other.total_vsz - self.vsz) / self.num_samples
            self.pmem += (other.total_pmem - self.pmem) / self.num_samples
            self.pcpu += (other.total_pcpu - self.pcpu) / self.num_samples

    @classmethod
    def from_sample(cls, sample: Sample) -> Averages:
        assert_num(
            sample.total_rss, sample.total_vsz, sample.total_pmem, sample.total_pcpu
        )
        return cls(
            rss=sample.total_rss,
            vsz=sample.total_vsz,
            pmem=sample.total_pmem,
            pcpu=sample.total_pcpu,
            num_samples=1,
        )


@dataclass
class Sample:
    stats: dict[int, ProcessStats] = field(default_factory=dict)
    averages: Averages = field(default_factory=Averages)
    total_rss: Optional[int] = None
    total_vsz: Optional[int] = None
    total_pmem: Optional[float] = None
    total_pcpu: Optional[float] = None
    timestamp: str = ""  # TS of last sample collected

    def add_pid(self, pid: int, stats: ProcessStats) -> None:
        # We do not calculate averages when we add a pid because we require all pids first
        assert (
            self.stats.get(pid) is None
        )  # add_pid should only be called when pid not in Sample
        self.total_rss = (self.total_rss or 0) + stats.rss
        self.total_vsz = (self.total_vsz or 0) + stats.vsz
        self.total_pmem = (self.total_pmem or 0.0) + stats.pmem
        self.total_pcpu = (self.total_pcpu or 0.0) + stats.pcpu
        self.stats[pid] = stats
        self.timestamp = max(self.timestamp, stats.timestamp)

    def aggregate(self: Sample, other: Sample) -> Sample:
        output = Sample()
        for pid in self.stats.keys() | other.stats.keys():
            if (mine := self.stats.get(pid)) is not None:
                if (theirs := other.stats.get(pid)) is not None:
                    output.add_pid(pid, mine.aggregate(theirs))
                else:
                    output.add_pid(pid, mine)
            else:
                output.add_pid(pid, other.stats[pid])
        assert other.total_pmem is not None
        assert other.total_pcpu is not None
        assert other.total_rss is not None
        assert other.total_vsz is not None
        output.total_pmem = max(self.total_pmem or 0.0, other.total_pmem)
        output.total_pcpu = max(self.total_pcpu or 0.0, other.total_pcpu)
        output.total_rss = max(self.total_rss or 0, other.total_rss)
        output.total_vsz = max(self.total_vsz or 0, other.total_vsz)
        output.averages = self.averages
        output.averages.update(other)
        return output

    def for_json(self) -> dict[str, Any]:
        d = {
            "timestamp": self.timestamp,
            "num_samples": self.averages.num_samples,
            "processes": {
                str(pid): stats.for_json() for pid, stats in self.stats.items()
            },
            "totals": {  # total of all processes during this sample
                "pmem": self.total_pmem,
                "pcpu": self.total_pcpu,
                "rss": self.total_rss,
                "vsz": self.total_vsz,
            },
            "averages": asdict(self.averages) if self.averages.num_samples >= 1 else {},
        }
        return d



"""Centralized JSON file type detection and loading for duct."""

from __future__ import annotations
import json
from typing import Any
from con_duct._constants import SUFFIXES

# Suffixes that use JSON Lines format
JSONL_SUFFIXES = (SUFFIXES["usage"], SUFFIXES["usage_legacy"])


def is_jsonl_file(path: str) -> bool:
    """Check if path is a JSON Lines file."""
    if any(path.endswith(s) for s in JSONL_SUFFIXES):
        return True
    return path.endswith(".jsonl")


def is_info_file(path: str) -> bool:
    """Check if path is a duct info file (standard JSON)."""
    return path.endswith(SUFFIXES["info"])


def load_usage_file(path: str) -> list[dict[str, Any]]:
    """Load a duct usage file (JSON Lines format)."""
    with open(path, "r") as f:
        return [json.loads(line) for line in f if line.strip()]


def load_info_file(path: str) -> dict[str, Any]:
    """Load a duct info file (standard JSON)."""
    with open(path, "r") as f:
        result: dict[str, Any] = json.load(f)
        return result



import argparse
import logging
import os
from pathlib import Path
import re
import sys
import textwrap
from typing import List, Optional
from con_duct import __version__
from con_duct._duct_main import DUCT_OUTPUT_PREFIX, EXECUTION_SUMMARY_FORMAT
from con_duct._duct_main import execute as duct_execute
from con_duct._models import Outputs, RecordTypes, SessionMode
from con_duct.ls import LS_FIELD_CHOICES, ls
from con_duct.plot import CPU_MODE_PS_PCPU, CPU_MODES, matplotlib_plot
from con_duct.pprint_json import pprint_json

# Default .env file search paths (in precedence order)
DEFAULT_CONFIG_PATHS_LIST = (
    "/etc/duct/.env",
    "${XDG_CONFIG_HOME:-~/.config}/duct/.env",
    ".duct/.env",
)
DEFAULT_CONFIG_PATHS = os.pathsep.join(DEFAULT_CONFIG_PATHS_LIST)


def load_duct_env_files() -> List[tuple[str, str]]:
    """Load environment variables from .env files in multiple locations.

    Searches for .env files specified in DUCT_CONFIG_PATHS (or DEFAULT_CONFIG_PATHS).
    Files are loaded in reverse order so later files override earlier ones.

    Environment variables already set in the environment will NOT be overridden
    by values from .env files, maintaining proper precedence:
    CLI args > explicit env vars > .env files > hardcoded defaults

    Gracefully handles missing python-dotenv package or missing .env files.

    Returns:
        List of (level_name, message) tuples for deferred logging.
    """
    log_buffer: List[tuple[str, str]] = []

    try:
        from dotenv import load_dotenv
    except ImportError:
        # python-dotenv not installed, skip .env file loading
        log_buffer.append(
            ("INFO", "python-dotenv not installed, skipping .env file loading")
        )
        return log_buffer

    config_paths_str = os.getenv("DUCT_CONFIG_PATHS", DEFAULT_CONFIG_PATHS)
    log_buffer.append(("DEBUG", f"Searching for .env files in: {config_paths_str}"))

    # Expand ${VAR:-default} syntax in the paths string ie ${XDG_CONFIG_HOME:-~/.config}
    def expand_var(match: re.Match) -> str:
        var_expr = match.group(1)
        if ":-" in var_expr:
            var_name, default = var_expr.split(":-", 1)
            return os.getenv(var_name, default)
        return os.getenv(var_expr, "")

    config_paths_str = re.sub(r"\$\{([^}]+)\}", expand_var, config_paths_str)
    search_paths = [
        val for p in config_paths_str.split(os.pathsep) if (val := p.strip())
    ]

    # Load in reverse order so later paths override earlier ones (once set, vars are skipped)
    loaded_count = 0
    for path in reversed(search_paths):
        expanded_path = Path(path).expanduser()
        if expanded_path.exists():
            log_buffer.append(("INFO", f"Loading .env file: {expanded_path}"))
            try:
                load_dotenv(expanded_path, override=False)
                loaded_count += 1
            except PermissionError as e:
                log_buffer.append(
                    ("WARNING", f"Cannot read .env file {expanded_path}: {e}")
                )
            except ValueError as e:
                log_buffer.append(
                    ("WARNING", f"Skipping malformed .env file {expanded_path}: {e}")
                )
        else:
            log_buffer.append(
                ("DEBUG", f".env file not found (skipping): {expanded_path}")
            )

    if loaded_count == 0:
        log_buffer.append(("INFO", "No .env files found"))

    return log_buffer


def _replay_early_logs(log_buffer: List[tuple[str, str]]) -> None:
    """Replay buffered log messages through the configured logger.

    Should be called after setup_logging() to ensure buffered messages from
    .env file loading are properly logged with the user's chosen log level.

    Args:
        log_buffer: List of (level_name, message) tuples to replay.
    """
    for level_name, message in log_buffer:
        lgr.log(getattr(logging, level_name), message)


lgr = logging.getLogger("con-duct")

# Format default config paths as a bulleted list for help text
_config_paths_list = "\n".join(f"    - {path}" for path in DEFAULT_CONFIG_PATHS_LIST)

ABOUT_DUCT = f"""
duct is a lightweight wrapper that collects execution data for an arbitrary
command. This command can be invoked as either 'duct' or 'con-duct run'.

Execution data includes execution time, system information, and resource usage
statistics of the command and all its child processes. It is intended to
simplify the problem of recording the resources necessary to execute a command,
particularly in an HPC environment.

Resource usage is determined by polling (at a sample-interval).
During execution, duct produces a JSON lines (see https://jsonlines.org) file
with one data point recorded for each report (at a report-interval).

limitations:
  Duct uses session id to track the command process and its children, so it
  cannot handle the situation where a process creates a new session.
  If a command spawns child processes, duct will collect data on them, but
  duct exits as soon as the primary process exits.

environment variables:
  Many duct options can be configured by environment variables (which are
  overridden by command line options).

  DUCT_LOG_LEVEL: see --log-level
  DUCT_OUTPUT_PREFIX: see --output-prefix
  DUCT_SUMMARY_FORMAT: see --summary-format
  DUCT_SAMPLE_INTERVAL: see --sample-interval
  DUCT_REPORT_INTERVAL: see --report-interval
  DUCT_CAPTURE_OUTPUTS: see --capture-outputs
  DUCT_MESSAGE: see --message
  DUCT_CONFIG_PATHS: paths to .env files separated by platform path separator
    (':' on Unix) (see below)

.env files:
  Environment variables can also be set via .env files. By default, duct
  searches the following locations (later files override earlier ones):

{_config_paths_list}

  Override the search paths by setting DUCT_CONFIG_PATHS with paths separated
  by the platform path separator (':' on Unix)
  (e.g., export DUCT_CONFIG_PATHS="/custom/path.env:~/.myduct.env").

  Example .env file content:
    # Set default log level
    DUCT_LOG_LEVEL=DEBUG

    # Configure intervals
    DUCT_SAMPLE_INTERVAL=2.0
    DUCT_REPORT_INTERVAL=120.0

    # Set default output location
    DUCT_OUTPUT_PREFIX=~/duct-logs/{{datetime}}-{{pid}}_

    # Add execution notes (multiline)
    DUCT_MESSAGE="Experiment run for paper revision
    Using updated dataset from 2025-10-30
    See lab notebook page 42 for details"

  Supported .env syntax (via python-dotenv):
    - KEY=value (basic assignment)
    - KEY="value with spaces" (quoted values)
    - KEY='single quotes' (single-quoted values)
    - # comments (hash comments)
    - Empty lines (ignored)
    - Multiline values (use quotes)

  Precedence (highest to lowest):
    1. Command line arguments
    2. Explicit environment variables
    3. .env file values (later paths override earlier paths)
    4. Hardcoded defaults

  Notes:
    - .env file support requires python-dotenv (pip install con-duct[all])
    - DUCT_CONFIG_PATHS cannot be set in .env files (must be set before loading)
    - Malformed, unreadable, or missing .env files are skipped and logged
"""


class CustomHelpFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """Override allows helptext to respect newlines in ABOUT_DUCT"""

    def _fill_text(self, text: str, width: int, _indent: str) -> str:
        return "\n".join([textwrap.fill(line, width) for line in text.splitlines()])


def _create_common_parser() -> argparse.ArgumentParser:
    """Create a parser with common arguments shared across all commands."""
    parser = argparse.ArgumentParser(add_help=False)  # help provided by child
    parser.add_argument(
        "-l",
        "--log-level",
        default=os.getenv("DUCT_LOG_LEVEL", "INFO").upper(),
        choices=("NONE", "CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"),
        type=str.upper,
        help="Level of log output to stderr, use NONE to entirely disable.",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="[deprecated, use log level NONE] Disable duct logging output (to stderr)",
    )
    return parser


def setup_logging(args: argparse.Namespace) -> None:
    """Configure logging based on parsed arguments.

    Handles both --log-level and --quiet flags, applying them consistently
    across all subcommands.
    """
    log_level = args.log_level

    # Handle deprecated --quiet flag
    if args.quiet:
        log_level = "NONE"

    # Special case: NONE means disable all logging
    if log_level == "NONE":
        logging.disable(logging.CRITICAL)
    else:
        logging.basicConfig(
            format="%(asctime)s [%(levelname)-8s] %(name)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
            level=log_level,
        )


def _create_run_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser for the 'run' command."""
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description=ABOUT_DUCT,
        formatter_class=CustomHelpFormatter,
        add_help=False,  # help provided by child
    )
    parser.add_argument(
        "command",
        metavar="command [command_args ...]",
        help="The command to execute, along with its arguments.",
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "command_args", nargs=argparse.REMAINDER, help="Arguments for the command."
    )
    parser.add_argument(
        "-p",
        "--output-prefix",
        type=str,
        default=DUCT_OUTPUT_PREFIX,
        help="File string format to be used as a prefix for the files -- the captured "
        "stdout and stderr and the resource usage logs. The understood variables are "
        "{datetime} and {pid}. "
        "Leading directories will be created if they do not exist. "
        "You can also provide value via DUCT_OUTPUT_PREFIX env variable. ",
    )
    parser.add_argument(
        "--summary-format",
        type=str,
        default=os.getenv("DUCT_SUMMARY_FORMAT", EXECUTION_SUMMARY_FORMAT),
        help="Output template to use when printing the summary following execution. "
        "Accepts custom conversion flags: "
        "!S: Converts filesizes to human readable units, green if measured, red if None. "
        "!E: Colors exit code, green if falsey, red if truthy, and red if None. "
        "!X: Colors green if truthy, red if falsey. "
        "!N: Colors green if not None, red if None",
    )
    parser.add_argument(
        "--colors",
        action="store_true",
        default=os.getenv("DUCT_COLORS", False),
        help="Use colors in duct output.",
    )
    parser.add_argument(
        "--clobber",
        action="store_true",
        help="Replace log files if they already exist.",
    )
    parser.add_argument(
        "--sample-interval",
        "--s-i",
        type=float,
        default=float(os.getenv("DUCT_SAMPLE_INTERVAL", "1.0")),
        help="Interval in seconds between status checks of the running process. "
        "Sample interval must be less than or equal to report interval, and it achieves the "
        "best results when sample is significantly less than the runtime of the process. "
        "Values below 1.0 behave erratically.",
    )
    parser.add_argument(
        "--report-interval",
        "--r-i",
        type=float,
        default=float(os.getenv("DUCT_REPORT_INTERVAL", "60.0")),
        help="Interval in seconds at which to report aggregated data.",
    )
    parser.add_argument(
        "--fail-time",
        "--f-t",
        type=float,
        default=float(os.getenv("DUCT_FAIL_TIME", "3.0")),
        help="If command fails in less than this specified time (seconds), duct would remove logs. "
        "Set to 0 if you would like to keep logs for a failing command regardless of its run time. "
        "Set to negative (e.g. -1) if you would like to not keep logs for any failing command.",
    )

    parser.add_argument(
        "-c",
        "--capture-outputs",
        default=os.getenv("DUCT_CAPTURE_OUTPUTS", "all"),
        choices=list(Outputs),
        type=Outputs,
        help="Record stdout, stderr, all, or none to log files. "
        "You can also provide value via DUCT_CAPTURE_OUTPUTS env variable.",
    )
    parser.add_argument(
        "-o",
        "--outputs",
        default="all",
        choices=list(Outputs),
        type=Outputs,
        help="Print stdout, stderr, all, or none to stdout/stderr respectively.",
    )
    parser.add_argument(
        "-t",
        "--record-types",
        default="all",
        choices=list(RecordTypes),
        type=RecordTypes,
        help="Record system-summary, processes-samples, or all",
    )
    parser.add_argument(
        "-m",
        "--message",
        type=str,
        default=os.getenv("DUCT_MESSAGE", ""),
        help="Record a descriptive message about the purpose of this execution. "
        "You can also provide value via DUCT_MESSAGE env variable.",
    )
    parser.add_argument(
        "--mode",
        default="new-session",
        choices=list(SessionMode),
        type=SessionMode,
        help="Session mode: 'new-session' creates a new session for the command (default), "
        "'current-session' tracks the current session instead of starting a new one. "
        "Useful for tracking slurm jobs or other commands that should run in the current session.",
    )
    return parser


def _create_pp_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser for the 'pp' command."""
    parser = argparse.ArgumentParser(
        add_help=False,  # help provided by child
    )
    parser.add_argument("file_path", help="JSON file to pretty print.")
    parser.add_argument(
        "-H",
        "--humanize",
        action="store_true",
        help="Convert numeric values to human-readable format",
    )
    return parser


def _create_plot_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser for the 'plot' command."""
    parser = argparse.ArgumentParser(
        add_help=False,  # help provided by child
    )
    parser.add_argument("file_path", help="duct-produced usage file.")
    parser.add_argument(
        "-o",
        "--output",
        help="Output path for the image file. If not specified, plot will be shown "
        "interactively (requires a display). Use this option in headless/server environments.",
        default=None,
    )
    parser.add_argument(
        "--min-ratio",
        type=float,
        default=3.0,
        help="Minimum ratio for axis unit selection (default: 3.0). Lower values use larger units sooner. "
        "Use -1 to always use base units (seconds, bytes).",
    )
    parser.add_argument(
        "--cpu",
        choices=CPU_MODES,
        default=CPU_MODE_PS_PCPU,
        help="Which CPU value to plot. 'ps-pcpu' uses the raw lifetime ratio "
        "from ps with no transformation. 'ps-cpu-timepoint' computes a "
        "delta-corrected time-point estimate from consecutive (pcpu, etime) "
        "pairs.",
    )
    return parser


def _create_ls_parser() -> argparse.ArgumentParser:
    """Create and configure the argument parser for the 'ls' command."""
    parser = argparse.ArgumentParser(
        add_help=False,  # help provided by child
    )
    parser.add_argument(
        "-f",
        "--format",
        choices=("auto", "pyout", "summaries", "json", "json_pp", "yaml"),
        default="auto",
        help="Output format. TODO Fixme. 'auto' chooses 'pyout' if pyout library is installed,"
        " 'summaries' otherwise.",
    )
    parser.add_argument(
        "-F",
        "--fields",
        nargs="+",
        metavar="FIELD",
        help=f"List of fields to show. Prefix is always included implicitly as the first field. "
        f"Available choices: {', '.join(sorted(LS_FIELD_CHOICES))}.",
        choices=LS_FIELD_CHOICES,
        default=[
            "command",
            "exit_code",
            "wall_clock_time",
            "peak_rss",
        ],
    )
    parser.add_argument(
        "--colors",
        action="store_true",
        default=os.getenv("DUCT_COLORS", False),
        help="Use colors in duct output.",
    )
    parser.add_argument(
        "paths",
        nargs="*",
        help="Path to duct report files, only `info.json` would be considered. "
        "If not provided, the program will glob for files that match DUCT_OUTPUT_PREFIX.",
    )
    parser.add_argument(
        "-e",
        "--eval-filter",
        help="Python expression to filter results based on available fields. "
        "The expression is evaluated for each entry, and only those that return True are included. "
        "See --fields for all supported fields. "
        "Example: --eval-filter \"filter_this=='yes'\" filters entries where 'filter_this' is 'yes'. "
        "You can use 're' for regex operations (e.g., --eval-filter \"re.search('2025.02.09.*', prefix)\").",
    )
    parser.add_argument(
        "--reverse",
        action="store_true",
        help="List entries in reverse order (most recent first).",
    )
    return parser


def run_command(args: argparse.Namespace) -> int:
    """Execute a command with duct monitoring."""
    kwargs = vars(args).copy()
    # Remove arguments that are not for duct_execute
    for key in ("func", "log_level", "quiet"):
        kwargs.pop(key, None)
    return duct_execute(**kwargs)


def execute(args: argparse.Namespace) -> int:
    """Execute the subcommand function and return its exit code."""
    result = args.func(args)
    if not isinstance(result, int):
        raise TypeError(
            f"Each con-duct subcommand must return an int returncode, got {type(result)}"
        )
    return result


def duct_entrypoint() -> None:
    """Entry point for the 'duct' command - delegates to 'con-duct run'."""
    os.execvp("con-duct", ["con-duct", "run"] + sys.argv[1:])


def main(argv: Optional[List[str]] = None) -> None:
    # Load .env files before parser creation so defaults pick up env vars
    env_log_buffer = load_duct_env_files()

    parser = argparse.ArgumentParser(
        prog="con-duct",
        description="A suite of commands to manage or manipulate con-duct logs.",
        usage="con-duct <command> [options]",
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    common_parser = _create_common_parser()
    subparsers = parser.add_subparsers(dest="command", help="Available subcommands")

    # Subcommand: run
    parser_run = subparsers.add_parser(
        "run",
        help="Execute a command with duct monitoring.",
        description=ABOUT_DUCT,
        parents=[common_parser, _create_run_parser()],
        formatter_class=CustomHelpFormatter,
        allow_abbrev=False,
        prog="con-duct run",
    )
    parser_run.set_defaults(func=run_command)

    # Subcommand: pp
    parser_pp = subparsers.add_parser(
        "pp",
        help="Pretty print a JSON log.",
        parents=[common_parser, _create_pp_parser()],
        prog="con-duct pp",
    )
    parser_pp.set_defaults(func=pprint_json)

    # Subcommand: plot
    parser_plot = subparsers.add_parser(
        "plot",
        help="Plot resource usage for an execution.",
        parents=[common_parser, _create_plot_parser()],
        prog="con-duct plot",
    )
    parser_plot.set_defaults(func=matplotlib_plot)

    parser_ls = subparsers.add_parser(
        "ls",
        help="Print execution information for all matching runs.",
        parents=[common_parser, _create_ls_parser()],
        prog="con-duct ls",
    )
    parser_ls.set_defaults(func=ls)

    args = parser.parse_args(argv)

    if args.command is None:
        parser.print_help()
        return

    setup_logging(args)
    _replay_early_logs(env_log_buffer)
    sys.exit(execute(args))


# To allow users or devs to invoke the `cli.py` directly
# instead of the entrypoint `con-duct`
# (used mainly by PyInstaller)
if __name__ == "__main__":
    main()



import argparse
import json
import logging
from pprint import pprint
from typing import Any
from con_duct._formatter import SummaryFormatter
from con_duct.json_utils import is_jsonl_file, load_info_file, load_usage_file

lgr = logging.getLogger(__name__)


def get_field_conversion_mapping() -> dict[str, str]:
    """
    Map field names to SummaryFormatter conversion types.
    """
    return {
        "average_pcpu": "P",
        "average_pmem": "P",
        "average_rss": "S",
        "average_vsz": "S",
        "peak_pcpu": "P",
        "peak_pmem": "P",
        "memory_total": "S",
        "peak_rss": "S",
        "peak_vsz": "S",
        "wall_clock_time": "T",
        "start_time": "D",
        "end_time": "D",
    }


def _apply_conversion(
    key: str, value: Any, field_mapping: dict[str, str], formatter: SummaryFormatter
) -> Any:
    """
    Apply SummaryFormatter conversion to a value based on field name.
    """
    if not isinstance(value, (int, float)):
        return value

    conversion = field_mapping.get(key.lower())
    if conversion:
        return formatter.convert_field(str(value), conversion)

    return value


def humanize_data(data: Any, formatter: SummaryFormatter) -> Any:
    """
    Recursively humanize numeric values using SummaryFormatter conversions.
    """
    field_mapping = get_field_conversion_mapping()

    if isinstance(data, dict):
        return {
            key: humanize_data(
                _apply_conversion(key, value, field_mapping, formatter), formatter
            )
            for key, value in data.items()
        }
    elif isinstance(data, list):
        return [humanize_data(item, formatter) for item in data]
    else:
        return data


def pprint_json(args: argparse.Namespace) -> int:
    """
    Prints the contents of a JSON file using pprint.

    Handles both standard JSON files and JSON Lines (usage.jsonl) files.
    """
    try:
        if is_jsonl_file(args.file_path):
            data: Any = load_usage_file(args.file_path)
        else:
            data = load_info_file(args.file_path)

        if args.humanize:
            formatter = SummaryFormatter()
            data = humanize_data(data, formatter)

        pprint(data)

    except FileNotFoundError:
        lgr.error("File not found: %s", args.file_path)
        return 1
    except json.JSONDecodeError as e:
        lgr.error("Error decoding JSON: %s", e)
        return 1

    return 0



"""Platform-specific process sampling for con-duct."""

from __future__ import annotations
from collections import Counter, deque
from datetime import datetime
import logging
import os
import platform
import subprocess
import sys
from typing import Callable, Optional
from con_duct._models import Averages, ProcessStats, Sample

SYSTEM = platform.system()

_SUPPORTED_SYSTEMS = {"Linux", "Darwin"}
if SYSTEM not in _SUPPORTED_SYSTEMS:
    sys.tracebacklimit = 0
    message = (
        f"`con_duct` does not currently support the detected operating system ({SYSTEM}).\n\n"
        "If you would like to request support, please open an issue at: "
        "https://github.com/con/duct/issues/new"
    )
    raise NotImplementedError(message)

lgr = logging.getLogger("con-duct")


def _get_sample_linux(session_id: int) -> Sample:
    sample = Sample()

    ps_command = [
        "ps",
        "-w",
        "-s",
        str(session_id),
        "-o",
        "pid,pcpu,pmem,rss,vsz,etime,stat,cmd",
    ]
    output = subprocess.check_output(ps_command, text=True)

    for line in output.splitlines()[1:]:
        if not line:
            continue

        pid, pcpu, pmem, rss_kib, vsz_kib, etime, stat, cmd = line.split(maxsplit=7)

        sample.add_pid(
            pid=int(pid),
            stats=ProcessStats(
                pcpu=float(pcpu),
                pmem=float(pmem),
                rss=int(rss_kib) * 1024,
                vsz=int(vsz_kib) * 1024,
                timestamp=datetime.now().astimezone().isoformat(),
                etime=etime,
                stat=Counter([stat]),
                cmd=cmd,
            ),
        )
    sample.averages = Averages.from_sample(sample=sample)
    return sample


def _try_to_get_sid(pid: int) -> int:
    """
    It is possible that the `pid` returned by the top `ps` call no longer exists at time of `getsid` request.
    """
    try:
        return os.getsid(pid)
    except ProcessLookupError as exc:
        lgr.debug(f"Error fetching session ID for PID {pid}: {str(exc)}")
        return -1


def _get_ps_lines_mac() -> list[str]:
    ps_command = [
        "ps",
        "-ax",
        "-o",
        "pid,pcpu,pmem,rss,vsz,etime,stat,args",
    ]
    output = subprocess.check_output(ps_command, text=True)

    lines = [line for line in output.splitlines()[1:] if line]
    return lines


def _add_pid_to_sample_from_line_mac(
    line: str, pid_to_matching_sid: dict[int, int], sample: Sample
) -> None:
    pid, pcpu, pmem, rss_kb, vsz_kb, etime, stat, cmd = line.split(maxsplit=7)

    if pid_to_matching_sid.get(int(pid)) is not None:
        sample.add_pid(
            pid=int(pid),
            stats=ProcessStats(
                pcpu=float(pcpu),
                pmem=float(pmem),
                rss=int(rss_kb) * 1024,
                vsz=int(vsz_kb) * 1024,
                timestamp=datetime.now().astimezone().isoformat(),
                etime=etime,
                stat=Counter([stat]),
                cmd=cmd,
            ),
        )


def _get_sample_mac(session_id: int) -> Optional[Sample]:
    sample = Sample()

    lines = _get_ps_lines_mac()
    pid_to_matching_sid = {
        pid: sid
        for line in lines
        if (sid := _try_to_get_sid(pid=(pid := int(line.split(maxsplit=1)[0]))))
        == session_id
    }

    if not pid_to_matching_sid:
        lgr.debug(f"No processes found for session ID {session_id}.")
        return None

    # collections.deque with maxlen=0 is used to approximate the
    # performance of list comprehension (superior to basic for-loop)
    # and also does not store `None` (or other) return values
    deque(
        (
            _add_pid_to_sample_from_line_mac(  # type: ignore[func-returns-value]
                line=line, pid_to_matching_sid=pid_to_matching_sid, sample=sample
            )
            for line in lines
        ),
        maxlen=0,
    )

    sample.averages = Averages.from_sample(sample=sample)
    return sample


_get_sample_per_system = {
    "Linux": _get_sample_linux,
    "Darwin": _get_sample_mac,
}
_get_sample: Callable[[int], Optional[Sample]] = _get_sample_per_system[SYSTEM]  # type: ignore[assignment]



"""Constants used throughout con-duct."""

__schema_version__ = "0.2.2"

ENV_PREFIXES = ("PBS_", "SLURM_", "OSG")
SUFFIXES = {
    "stdout": "stdout",
    "stderr": "stderr",
    "usage": "usage.jsonl",
    "usage_legacy": "usage.json",
    "info": "info.json",
}
